# Orca
#
# Copyright 2026 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-lines
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-arguments,
# pylint: disable=too-many-positional-arguments
# pylint: disable=too-many-return-statements

"""Module for document-related presentation and navigation settings."""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Any

import gi

gi.require_version("Atspi", "2.0")
from gi.repository import Atspi  # pylint: disable=no-name-in-module

from . import (
    caret_navigator,
    command_manager,
    dbus_service,
    debug,
    document_presenter_command_definitions,
    focus_manager,
    gsettings_registry,
    guilabels,
    input_event,
    input_event_manager,
    math_navigator,
    messages,
    presentation_manager,
    script_manager,
    structural_navigator,
    table_navigator,
)
from .ax_object import AXObject
from .ax_text import AXText
from .ax_utilities import AXUtilities
from .ax_utilities_math import AXUtilitiesMath
from .extension import Extension

if TYPE_CHECKING:
    from collections.abc import Callable

    from .command import Command
    from .dbus_service import UInt32
    from .document_presenter_preferences_grid import DocumentPreferencesGrid
    from .scripts import default


@gsettings_registry.get_registry().gsettings_enum(
    "org.gnome.Orca.FindResultsVerbosity",
    values={"none": 0, "if-line-changed": 1, "all": 2},
)
class FindResultsVerbosity(Enum):
    """Find results verbosity level enumeration."""

    NONE = 0
    IF_LINE_CHANGED = 1
    ALL = 2

    @property
    def string_name(self) -> str:
        """Returns the lowercase string name for this enum value."""

        return self.name.lower().replace("_", "-")


@dataclass
class _AppModeState:
    """Tracks focus/browse mode state for a specific application."""

    in_focus_mode: bool = True
    focus_mode_is_sticky: bool = False
    browse_mode_is_sticky: bool = False
    user_has_overridden_auto_sticky_focus_mode: bool = False
    document_with_last_manual_mode_change: int | None = None


@gsettings_registry.get_registry().gsettings_schema("org.gnome.Orca.Document", name="document")
class DocumentPresenter(Extension):
    """Manages document-related presentation and navigation settings."""

    GROUP_LABEL = guilabels.KB_GROUP_DOCUMENTS

    _SCHEMA = "document"
    KEY_NATIVE_NAV_TRIGGERS_FOCUS_MODE = "native-nav-triggers-focus-mode"
    KEY_AUTO_STICKY_FOCUS_MODE = "auto-sticky-focus-mode"
    KEY_SAY_ALL_ON_LOAD = "say-all-on-load"
    KEY_PAGE_SUMMARY_ON_LOAD = "page-summary-on-load"
    KEY_FIND_RESULTS_VERBOSITY = "find-results-verbosity"
    KEY_FIND_RESULTS_MINIMUM_LENGTH = "find-results-minimum-length"

    def _get_setting(self, key: str, gtype: str, default: Any) -> Any:
        """Returns the dconf value for key, or default if not in dconf."""

        return gsettings_registry.get_registry().layered_lookup(
            self._SCHEMA,
            key,
            gtype,
            default=default,
        )

    def __init__(self) -> None:
        self._made_find_announcement = False
        self._app_states: dict[int, _AppModeState] = {}
        super().__init__()

    def _register_commands(self) -> None:
        """Registers commands and sets up exclusive groups."""

        super()._register_commands()

        manager = command_manager.get_manager()
        manager.add_exclusive_groups(
            guilabels.KB_GROUP_MATH_NAVIGATION,
            guilabels.KB_GROUP_CARET_NAVIGATION,
        )
        manager.add_exclusive_groups(
            guilabels.KB_GROUP_MATH_NAVIGATION,
            guilabels.KB_GROUP_STRUCTURAL_NAVIGATION,
        )
        manager.add_exclusive_groups(
            guilabels.KB_GROUP_MATH_NAVIGATION,
            guilabels.KB_GROUP_TABLE_NAVIGATION,
        )

    def _get_commands(self) -> list[Command]:
        """Returns commands for registration."""

        return document_presenter_command_definitions.get_commands(self)

    def _get_state_for_app(self, app: Atspi.Accessible | None) -> _AppModeState:
        """Returns the mode state for the given app, creating if needed."""

        if app is None:
            return _AppModeState()

        app_hash = hash(app)
        if app_hash not in self._app_states:
            self._app_states[app_hash] = _AppModeState()
        return self._app_states[app_hash]

    def _get_current_app(self) -> Atspi.Accessible | None:
        """Returns the current application from the active script."""

        script = script_manager.get_manager().get_active_script()
        if script is None:
            return None
        return script.app

    def _is_likely_electron_app(self, app: Atspi.Accessible | None) -> bool:
        """Returns True if app is likely an Electron app (Chromium-based, not a browser)."""

        if app is None:
            return False

        toolkit = AXObject.get_toolkit_name(app).lower()
        if toolkit != "chromium":
            return False

        app_name = AXObject.get_name(app).lower()
        known_browsers = ("brave", "chromium", "edge", "chrome", "opera", "vivaldi")
        result = not any(browser in app_name for browser in known_browsers)
        tokens = ["DOCUMENT PRESENTER:", app, "is likely Electron app:", result]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return result

    def _is_top_level_web_app(self, script: default.Script, obj: Atspi.Accessible | None) -> bool:
        """Returns True if obj is in a top-level web application (e.g., Google Docs)."""

        if obj is None:
            return False

        document = script.utilities.active_document()
        if document is None:
            return False

        if not AXUtilities.is_embedded(document):
            return False

        uri = AXUtilities.get_uri(document)
        result = bool(uri and uri.startswith("http"))
        tokens = ["DOCUMENT PRESENTER:", document, "is top-level web app:", result, ". URI:", uri]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return result

    def in_focus_mode(self, app: Atspi.Accessible | None = None) -> bool:
        """Returns True if focus mode is active for the given app."""

        if app is None:
            app = self._get_current_app()
        if app is None:
            return False
        app_hash = hash(app)
        if app_hash not in self._app_states:
            return False
        return self._app_states[app_hash].in_focus_mode

    def user_has_overridden_auto_sticky_focus_mode(
        self, app: Atspi.Accessible | None = None
    ) -> bool:
        """Returns True if the user has changed the mode Orca automatically made sticky."""

        if app is None:
            app = self._get_current_app()
        if app is None:
            return False
        app_hash = hash(app)
        if app_hash not in self._app_states:
            return False
        return self._app_states[app_hash].user_has_overridden_auto_sticky_focus_mode

    def set_user_has_overridden_auto_sticky_focus_mode(
        self, app: Atspi.Accessible | None, overridden: bool
    ) -> None:
        """Sets whether the user has changed the mode Orca automatically made sticky."""

        self._get_state_for_app(app).user_has_overridden_auto_sticky_focus_mode = overridden

    def user_last_changed_mode_in_document(
        self, script: default.Script, obj: Atspi.Accessible | None
    ) -> bool:
        """Returns True if the user last changed the mode in the document containing obj."""

        state = self._get_state_for_app(script.app)
        if state.document_with_last_manual_mode_change is None:
            return False

        document = script.utilities.get_top_level_document_for_object(obj)
        if document is None:
            return False
        return state.document_with_last_manual_mode_change == hash(document)

    def set_document_with_last_manual_mode_change(
        self, script: default.Script, obj: Atspi.Accessible | None
    ) -> None:
        """Records the document containing obj as where the user last changed the mode."""

        document = script.utilities.get_top_level_document_for_object(obj)
        state = self._get_state_for_app(script.app)
        state.document_with_last_manual_mode_change = None if document is None else hash(document)

    def clear_document_with_last_manual_mode_change(self, app: Atspi.Accessible | None) -> None:
        """Clears the record of the document in which the user last changed the mode."""

        self._get_state_for_app(app).document_with_last_manual_mode_change = None

    def focus_mode_is_sticky(self, app: Atspi.Accessible | None = None) -> bool:
        """Returns True if focus mode is sticky for the given app."""

        if app is None:
            app = self._get_current_app()
        if app is None:
            return False
        app_hash = hash(app)
        if app_hash not in self._app_states:
            return False
        return self._app_states[app_hash].focus_mode_is_sticky

    def browse_mode_is_sticky(self, app: Atspi.Accessible | None = None) -> bool:
        """Returns True if browse mode is sticky for the given app."""

        if app is None:
            app = self._get_current_app()
        if app is None:
            return False
        app_hash = hash(app)
        if app_hash not in self._app_states:
            return False
        return self._app_states[app_hash].browse_mode_is_sticky

    def _set_presentation_mode(
        self,
        script: default.Script,
        use_focus_mode: bool,
        obj: Atspi.Accessible | None = None,
        document: Atspi.Accessible | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Sets the presentation mode to focus or browse mode."""

        tokens = [
            "DOCUMENT PRESENTER: set_presentation_mode. Use focus mode:",
            use_focus_mode,
            ",",
            obj,
            "in",
            document,
            "notify user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if obj is not None and AXObject.is_dead(obj):
            obj = None

        if not script.utilities.in_document_content(obj):
            if notify_user:
                presentation_manager.get_manager().present_message(messages.DOCUMENT_NOT_IN_A)
            return False

        has_state = self.has_state_for_app(script.app)
        in_focus_mode = self.in_focus_mode(script.app)
        if has_state and in_focus_mode == use_focus_mode:
            msg = "DOCUMENT PRESENTER: Presentation mode already set."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        obj, _offset = script.utilities.get_caret_context(document)

        if in_focus_mode and not use_focus_mode:
            parent = AXObject.get_parent(obj)
            if AXUtilities.is_list_box(parent):
                script.utilities.set_caret_context(parent, -1)
            elif AXUtilities.is_menu(parent):
                script.utilities.set_caret_context(AXObject.get_parent(parent), -1)

        if not in_focus_mode and use_focus_mode:
            if (
                caret_navigator.get_navigator().last_input_event_was_navigation_command()
                or structural_navigator.get_navigator().last_input_event_was_navigation_command()
                or table_navigator.get_navigator().last_input_event_was_navigation_command()
            ):
                AXObject.grab_focus(obj)

        if notify_user:
            msg = messages.MODE_FOCUS if use_focus_mode else messages.MODE_BROWSE
            presentation_manager.get_manager().present_message(msg)

        state = self._get_state_for_app(script.app)
        state.in_focus_mode = use_focus_mode
        state.focus_mode_is_sticky = False
        state.browse_mode_is_sticky = False

        reason = "setting presentation mode"
        if not use_focus_mode:
            self._enable_document_navigators(script, reason)

        self.suspend_navigators(script, use_focus_mode, reason)
        return True

    def suspend_navigators(self, script: default.Script, suspended: bool, reason: str) -> bool:
        """Suspends or unsuspends navigation commands. Returns True if state changed."""

        if not suspended and math_navigator.get_navigator().is_active():
            msg = "DOCUMENT PRESENTER: Not unsuspending navigators: math navigation is active."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        caret_navigator.get_navigator().suspend_commands(script, suspended, reason)
        structural_navigator.get_navigator().suspend_commands(script, suspended, reason)
        return True

    def enter_math_navigation(self, script: default.Script, obj: Atspi.Accessible) -> bool:
        """Enters math navigation mode for the given math object."""

        if not math_navigator.get_navigator().enter_math_mode(obj):
            return False

        self.suspend_navigators(script, True, "math navigation")
        math_navigator.get_navigator().suspend_commands(script, False, "math navigation")
        return True

    def exit_math_navigation(self, script: default.Script) -> None:
        """Exits math navigation mode and restores normal navigation."""

        nav = math_navigator.get_navigator()
        if nav.is_active():
            nav.reset(script)
        nav.suspend_commands(script, True, "leaving math navigation")
        self.suspend_navigators(script, False, "leaving math navigation")

    def _enable_document_navigators(self, script: default.Script, reason: str) -> None:
        """Enables document navigators for the given script."""

        tokens = ["DOCUMENT PRESENTER: _enable_document_navigators. Reason:", reason]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        structural_navigator.get_navigator().set_mode(
            script,
            structural_navigator.NavigationMode.DOCUMENT,
        )
        caret_navigator.get_navigator().set_enabled_for_script(script, True)

    def _is_focus_mode_widget_by_state(self, obj: Atspi.Accessible) -> tuple[bool, str]:
        """Returns (True, reason) if obj's state makes it a focus mode widget."""

        if AXUtilities.is_editable(obj):
            return True, "it's editable"

        if (
            AXUtilities.is_expandable(obj)
            and AXUtilities.is_focusable(obj)
            and not AXUtilities.is_link(obj)
        ):
            return True, "it's expandable and focusable"

        return False, ""

    def _is_focus_mode_widget_by_role(
        self,
        script: default.Script,
        obj: Atspi.Accessible,
    ) -> tuple[bool | None, str]:
        """Returns (True/False, reason) if role determines focus mode, or (None, '') if unclear."""

        always_focus_mode_roles = [
            Atspi.Role.COMBO_BOX,
            Atspi.Role.ENTRY,
            Atspi.Role.LIST_BOX,
            Atspi.Role.MENU,
            Atspi.Role.MENU_ITEM,
            Atspi.Role.CHECK_MENU_ITEM,
            Atspi.Role.RADIO_MENU_ITEM,
            Atspi.Role.PAGE_TAB,
            Atspi.Role.PASSWORD_TEXT,
            Atspi.Role.PROGRESS_BAR,
            Atspi.Role.SLIDER,
            Atspi.Role.SPIN_BUTTON,
            Atspi.Role.TOOL_BAR,
            Atspi.Role.TREE_ITEM,
            Atspi.Role.TREE_TABLE,
            Atspi.Role.TREE,
        ]

        role = AXObject.get_role(obj)
        if role in always_focus_mode_roles:
            return True, "due to its role"

        if role in (Atspi.Role.TABLE_CELL, Atspi.Role.TABLE):
            if AXUtilities.is_layout_table(AXUtilities.get_table(obj)):
                return False, "it's layout only"
            if AXUtilities.is_pdf(script.utilities.get_document_for_object(obj)):
                return False, "it's in a PDF"

        if AXUtilities.is_list_box_item(obj, role):
            return True, "it's a listbox item"

        if AXUtilities.is_button_with_popup(obj, role):
            return True, "it's a button with popup"

        if role == Atspi.Role.EMBEDDED:
            if script.utilities.has_name_and_action_and_no_useful_children(obj):
                return False, "has name and action and no useful children"
            return True, "based on presumed functionality"

        if role == Atspi.Role.TABLE and AXUtilities.is_grid(obj, role):
            return True, "it's a grid"

        return None, ""

    def _is_focus_mode_widget_by_ancestry(
        self,
        script: default.Script,
        obj: Atspi.Accessible,
    ) -> tuple[bool, str]:
        """Returns (True, reason) if obj's ancestry makes it a focus mode widget."""

        if AXUtilities.is_grid_descendant(obj):
            return True, "it's a grid descendant"
        if AXUtilities.is_menu_descendant(obj):
            return True, "it's a menu descendant"
        if AXUtilities.is_tool_bar_descendant(obj):
            return True, "it's a toolbar descendant"
        if script.utilities.is_content_editable_with_embedded_objects(obj):
            return True, "it's content editable"

        return False, ""

    def is_focus_mode_widget(self, script: default.Script, obj: Atspi.Accessible) -> bool:
        """Returns True if obj should be interacted with in focus mode."""

        result, reason = self._is_focus_mode_widget_by_state(obj)
        if not result:
            role_result, reason = self._is_focus_mode_widget_by_role(script, obj)
            if role_result is None:
                result, reason = self._is_focus_mode_widget_by_ancestry(script, obj)
            else:
                result = role_result

        prefix = "is" if result else "is not"
        if reason:
            tokens = ["DOCUMENT PRESENTER:", obj, prefix, "focus mode widget:", reason]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        return result

    @dbus_service.command
    def enable_sticky_browse_mode(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Enables sticky browse mode."""

        tokens = ["DOCUMENT PRESENTER: enable_sticky_browse_mode(", event, ",", notify_user, ")"]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if not script.utilities.in_document_content():
            if notify_user:
                presentation_manager.get_manager().present_message(messages.DOCUMENT_NOT_IN_A)
            return True

        state = self._get_state_for_app(script.app)

        if not state.browse_mode_is_sticky or notify_user:
            presentation_manager.get_manager().present_message(messages.MODE_BROWSE_IS_STICKY)

        state.in_focus_mode = False
        state.focus_mode_is_sticky = False
        state.browse_mode_is_sticky = True
        if event is not None:
            state.user_has_overridden_auto_sticky_focus_mode = True

        reason = "enable sticky browse mode"
        self.suspend_navigators(script, state.in_focus_mode, reason)
        return True

    @dbus_service.command
    def enable_sticky_focus_mode(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Enables sticky focus mode."""

        tokens = ["DOCUMENT PRESENTER: enable_sticky_focus_mode(", event, ",", notify_user, ")"]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if not script.utilities.in_document_content():
            if notify_user:
                presentation_manager.get_manager().present_message(messages.DOCUMENT_NOT_IN_A)
            return True

        state = self._get_state_for_app(script.app)

        if not state.focus_mode_is_sticky or notify_user:
            presentation_manager.get_manager().present_message(messages.MODE_FOCUS_IS_STICKY)

        state.in_focus_mode = True
        state.focus_mode_is_sticky = True
        state.browse_mode_is_sticky = False
        if event is not None:
            state.user_has_overridden_auto_sticky_focus_mode = True

        reason = "enable sticky focus mode"
        self.suspend_navigators(script, state.in_focus_mode, reason)
        return True

    @dbus_service.command
    def toggle_presentation_mode(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        document: Atspi.Accessible | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Switches between browse mode and focus mode (user-initiated)."""

        if not script.utilities.in_document_content():
            if notify_user:
                presentation_manager.get_manager().present_message(messages.DOCUMENT_NOT_IN_A)
            return True

        use_focus = not self.in_focus_mode(script.app)
        obj, _offset = script.utilities.get_caret_context(document)
        if event is not None and use_focus:
            AXObject.grab_focus(obj)

        self._set_presentation_mode(
            script,
            use_focus,
            obj=obj,
            document=document,
            notify_user=notify_user,
        )
        if event is not None:
            self.set_user_has_overridden_auto_sticky_focus_mode(script.app, True)
            self.set_document_with_last_manual_mode_change(script, obj)
        return True

    @dbus_service.getter
    def get_in_focus_mode(self) -> bool:
        """Returns True if focus mode is active (web content only)."""

        if script := script_manager.get_manager().get_active_script():
            return self.in_focus_mode(script.app)
        return False

    @dbus_service.getter
    def get_focus_mode_is_sticky(self) -> bool:
        """Returns True if focus mode is active and 'sticky' (web content only)."""

        if script := script_manager.get_manager().get_active_script():
            return self.focus_mode_is_sticky(script.app)
        return False

    @dbus_service.getter
    def get_browse_mode_is_sticky(self) -> bool:
        """Returns True if browse mode is active and 'sticky' (web content only)."""

        if script := script_manager.get_manager().get_active_script():
            return self.browse_mode_is_sticky(script.app)
        return False

    def restore_mode_for_script(self, script: default.Script) -> None:
        """Restores navigator suspension state when a script is activated."""

        if script.app is None:
            return

        app_hash = hash(script.app)
        if app_hash not in self._app_states:
            return

        state = self._app_states[app_hash]

        # When restoring browse mode, also re-enable the navigators.
        # This is needed because another script (e.g. file browser) may have
        # disabled them via set_mode(OFF) while it was active.
        reason = "restoring mode state for activated script"
        if not state.in_focus_mode:
            self._enable_document_navigators(script, reason)

        self.suspend_navigators(script, state.in_focus_mode, reason)

        tokens = [
            "DOCUMENT PRESENTER: Restored mode for",
            script.app,
            ". Focus mode:",
            state.in_focus_mode,
            "Focus sticky:",
            state.focus_mode_is_sticky,
            "Browse sticky:",
            state.browse_mode_is_sticky,
            "User overrode auto sticky focus mode:",
            state.user_has_overridden_auto_sticky_focus_mode,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

    def has_state_for_app(self, app: Atspi.Accessible | None) -> bool:
        """Returns True if mode state exists for the given app."""

        if app is None:
            return False
        return hash(app) in self._app_states

    def clear_state_for_app(self, app: Atspi.Accessible | None) -> None:
        """Clears mode state when an app is closed."""

        if app is None:
            return
        app_hash = hash(app)
        if app_hash in self._app_states:
            del self._app_states[app_hash]
            tokens = ["DOCUMENT PRESENTER: Cleared state for", app]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)

    def _handle_entering_document(
        self,
        script: default.Script,
        new_focus: Atspi.Accessible,
        old_focus: Atspi.Accessible | None,
    ) -> bool:
        """Handles mode/navigator setup when entering a document from outside."""

        if self.focus_mode_is_sticky(script.app):
            presentation_manager.get_manager().present_message(messages.MODE_FOCUS_IS_STICKY)
            reason = "locus of focus now in document, focus mode is sticky"
            self.suspend_navigators(script, True, reason)
            return True

        if self.browse_mode_is_sticky(script.app):
            presentation_manager.get_manager().present_message(messages.MODE_BROWSE_IS_STICKY)
            reason = "locus of focus now in document, browse mode is sticky"
            self._enable_document_navigators(script, reason)
            self.suspend_navigators(script, False, reason)
            return True

        # Only do app-type detection if setting is enabled and user hasn't explicitly
        # toggled mode. This allows the user to escape auto-enabled sticky focus mode.
        if (
            self.get_auto_sticky_focus_mode_for_web_apps()
            and not self.user_has_overridden_auto_sticky_focus_mode(script.app)
        ):
            if self._is_likely_electron_app(script.app):
                msg = "DOCUMENT PRESENTER: Electron app detected, enabling sticky focus mode"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                self.enable_sticky_focus_mode(script, notify_user=True)
                return True

            if self._is_top_level_web_app(script, new_focus):
                msg = "DOCUMENT PRESENTER: Top-level web app detected, enabling sticky focus mode"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                self.enable_sticky_focus_mode(script, notify_user=True)
                return True

        use_focus = self.use_focus_mode(new_focus, old_focus)
        reason = "entering document"
        if not use_focus:
            self._enable_document_navigators(script, reason)

        state = self._get_state_for_app(script.app)
        state.in_focus_mode = use_focus
        self.suspend_navigators(script, use_focus, reason)

        if use_focus:
            presentation_manager.get_manager().present_message(messages.MODE_FOCUS)
        else:
            presentation_manager.get_manager().present_message(messages.MODE_BROWSE)

        return True

    def update_mode_if_needed(
        self,
        script: default.Script,
        old_focus: Atspi.Accessible | None,
        new_focus: Atspi.Accessible | None,
    ) -> bool:
        """Updates focus/browse mode based on a focus change. Returns True if handled."""

        old_doc = script.utilities.get_top_level_document_for_object(old_focus)
        new_doc = script.utilities.get_top_level_document_for_object(new_focus)

        tokens = [
            "DOCUMENT PRESENTER: Updating mode for focus change.",
            "Old focus:",
            old_focus,
            "old doc:",
            old_doc,
            "New focus:",
            new_focus,
            "new doc:",
            new_doc,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if new_doc is None:
            self.reset_find_announcement_state()
            if math_navigator.get_navigator().is_active():
                self.exit_math_navigation(script)
            reason = "locus of focus no longer in document"
            self.suspend_navigators(script, False, reason)
            structural_navigator.get_navigator().set_mode(
                script,
                structural_navigator.NavigationMode.OFF,
            )
            caret_navigator.get_navigator().set_enabled_for_script(script, False)
            return True

        if math_navigator.get_navigator().is_active() and AXUtilitiesMath.find_math_root(new_focus):
            return True

        if old_doc is None and not focus_manager.get_manager().old_focus_was_dead():
            return self._handle_entering_document(script, new_focus, old_focus)

        # Focus change within document
        if self.focus_mode_is_sticky(script.app) or self.browse_mode_is_sticky(script.app):
            return False

        use_focus = self.use_focus_mode(new_focus, old_focus)
        if use_focus != self.in_focus_mode(script.app):
            msg = "DOCUMENT PRESENTER: Mode changed automatically. Clearing the manual change."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self.clear_document_with_last_manual_mode_change(script.app)
        self._set_presentation_mode(script, use_focus, obj=new_focus, document=new_doc)
        return True

    def reset_find_announcement_state(self) -> None:
        """Resets the find announcement state."""

        self._made_find_announcement = False

    @gsettings_registry.get_registry().gsetting(
        key=KEY_NATIVE_NAV_TRIGGERS_FOCUS_MODE,
        schema="document",
        gtype="b",
        default=True,
        summary="Native navigation triggers focus mode",
        migration_key="nativeNavTriggersFocusMode",
    )
    @dbus_service.getter
    def get_native_nav_triggers_focus_mode(self) -> bool:
        """Returns whether native navigation triggers focus mode."""

        return self._get_setting(self.KEY_NATIVE_NAV_TRIGGERS_FOCUS_MODE, "b", True)

    @dbus_service.setter
    def set_native_nav_triggers_focus_mode(self, value: bool) -> bool:
        """Sets whether native navigation triggers focus mode."""

        if self.get_native_nav_triggers_focus_mode() == value:
            return True

        tokens = ["DOCUMENT PRESENTER: Setting native nav triggers focus mode to", value, "."]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_NATIVE_NAV_TRIGGERS_FOCUS_MODE,
            value,
        )
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_AUTO_STICKY_FOCUS_MODE,
        schema="document",
        gtype="b",
        default=True,
        summary="Auto-detect sticky focus mode for web apps",
        migration_key="autoStickyFocusModeForWebApps",
    )
    @dbus_service.getter
    def get_auto_sticky_focus_mode_for_web_apps(self) -> bool:
        """Returns whether to auto-detect web apps and enable sticky focus mode."""

        return self._get_setting(self.KEY_AUTO_STICKY_FOCUS_MODE, "b", True)

    @dbus_service.setter
    def set_auto_sticky_focus_mode_for_web_apps(self, value: bool) -> bool:
        """Sets whether to auto-detect web apps and enable sticky focus mode."""

        if self.get_auto_sticky_focus_mode_for_web_apps() == value:
            return True

        tokens = ["DOCUMENT PRESENTER: Setting auto sticky focus mode for web apps to", value, "."]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_AUTO_STICKY_FOCUS_MODE,
            value,
        )
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_SAY_ALL_ON_LOAD,
        schema="document",
        gtype="b",
        default=True,
        summary="Perform say all when document loads",
        migration_key="sayAllOnLoad",
    )
    @dbus_service.getter
    def get_say_all_on_load(self) -> bool:
        """Returns whether to perform say all when a document loads."""

        return self._get_setting(self.KEY_SAY_ALL_ON_LOAD, "b", True)

    @dbus_service.setter
    def set_say_all_on_load(self, value: bool) -> bool:
        """Sets whether to perform say all when a document loads."""

        if self.get_say_all_on_load() == value:
            return True

        tokens = ["DOCUMENT PRESENTER: Setting say all on load to", value, "."]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA, self.KEY_SAY_ALL_ON_LOAD, value
        )
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_PAGE_SUMMARY_ON_LOAD,
        schema="document",
        gtype="b",
        default=True,
        summary="Present page summary when document loads",
        migration_key="pageSummaryOnLoad",
    )
    @dbus_service.getter
    def get_page_summary_on_load(self) -> bool:
        """Returns whether to present a page summary when a document loads."""

        return self._get_setting(self.KEY_PAGE_SUMMARY_ON_LOAD, "b", True)

    @dbus_service.setter
    def set_page_summary_on_load(self, value: bool) -> bool:
        """Sets whether to present a page summary when a document loads."""

        if self.get_page_summary_on_load() == value:
            return True

        tokens = ["DOCUMENT PRESENTER: Setting page summary on load to", value, "."]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_PAGE_SUMMARY_ON_LOAD,
            value,
        )
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_FIND_RESULTS_VERBOSITY,
        schema="document",
        genum="org.gnome.Orca.FindResultsVerbosity",
        default="all",
        summary="Find results verbosity (none, if-line-changed, all)",
        migration_key="findResultsVerbosity",
    )
    def get_find_results_verbosity_name(self) -> str:
        """Returns the find results verbosity level as a string name."""

        return gsettings_registry.get_registry().layered_lookup(
            self._SCHEMA,
            self.KEY_FIND_RESULTS_VERBOSITY,
            "",
            genum="org.gnome.Orca.FindResultsVerbosity",
            default="all",
        )

    @dbus_service.getter
    def get_speak_find_results(self) -> bool:
        """Returns whether to speak find results."""

        return self.get_find_results_verbosity_name() != "none"

    @dbus_service.setter
    def set_speak_find_results(self, value: bool) -> bool:
        """Sets whether to speak find results."""

        if self.get_speak_find_results() == value:
            return True

        tokens = ["DOCUMENT PRESENTER: Setting speak find results to", value, "."]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        name = "all" if value else "none"
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_FIND_RESULTS_VERBOSITY,
            name,
        )
        return True

    @dbus_service.getter
    def get_only_speak_changed_lines(self) -> bool:
        """Returns whether to only speak changed lines during find."""

        return self.get_find_results_verbosity_name() == "if-line-changed"

    @dbus_service.setter
    def set_only_speak_changed_lines(self, value: bool) -> bool:
        """Sets whether to only speak changed lines during find."""

        if self.get_only_speak_changed_lines() == value:
            return True

        tokens = ["DOCUMENT PRESENTER: Setting only speak changed lines to", value, "."]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        name = "if-line-changed" if value else "all"
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_FIND_RESULTS_VERBOSITY,
            name,
        )
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_FIND_RESULTS_MINIMUM_LENGTH,
        schema="document",
        gtype="i",
        default=4,
        summary="Minimum length for find results to be spoken",
        migration_key="findResultsMinimumLength",
    )
    @dbus_service.getter
    def get_find_results_minimum_length(self) -> UInt32:
        """Returns the minimum length for find results to be spoken."""

        return self._get_setting(self.KEY_FIND_RESULTS_MINIMUM_LENGTH, "i", 4)

    @dbus_service.setter
    def set_find_results_minimum_length(self, value: UInt32) -> bool:
        """Sets the minimum length for find results to be spoken."""

        if self.get_find_results_minimum_length() == value:
            return True

        tokens = ["DOCUMENT PRESENTER: Setting find results minimum length to", value, "."]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_FIND_RESULTS_MINIMUM_LENGTH,
            value,
        )
        return True

    def _restrict_find_results_to_cell(
        self,
        obj: Atspi.Accessible,
        contents: list[tuple[Atspi.Accessible, int, int, str]],
    ) -> list[tuple[Atspi.Accessible, int, int, str]]:
        """Limits find-result contents to obj's table cell rather than the whole row."""

        cell = AXUtilities.find_ancestor_inclusive(obj, AXUtilities.is_table_cell)
        if cell is None:
            return contents

        return [
            content
            for content in contents
            if AXUtilities.find_ancestor_inclusive(content[0], AXUtilities.is_table_cell) == cell
        ]

    # pylint: disable-next=too-many-locals
    def present_find_results(self, obj: Atspi.Accessible, offset: int) -> bool:
        """Presents find results if appropriate based on settings. Returns True if presented."""

        if not self.get_speak_find_results():
            msg = "DOCUMENT PRESENTER: Speaking find results is disabled."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        script = script_manager.get_manager().get_active_script()
        if script is None:
            msg = "DOCUMENT PRESENTER: No active script for find results presentation."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        document = script.utilities.get_document_for_object(obj)
        if document is None:
            msg = "DOCUMENT PRESENTER: No document for find results presentation."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        start = AXUtilities.get_selection_start_offset(obj)
        if start < 0:
            msg = "DOCUMENT PRESENTER: Invalid selection start for find results presentation."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        tokens = [
            "DOCUMENT PRESENTER: Find results",
            obj,
            "offset:",
            offset,
            ", selection start offset:",
            start,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        offset = max(offset, start)
        context = script.utilities.get_caret_context(document)
        script.utilities.set_caret_context(obj, offset, document=document)

        focus = focus_manager.get_manager().get_locus_of_focus()
        manager = input_event_manager.get_manager()
        user_is_editing_search_term = (
            manager.last_event_was_printable_key()
            or manager.last_event_was_backspace()
            or manager.last_event_was_delete()
        ) and AXUtilities.is_editable(focus)
        end = AXUtilities.get_selection_end_offset(obj)
        if end - start < self.get_find_results_minimum_length() and user_is_editing_search_term:
            msg = "DOCUMENT PRESENTER: Search term is shorter than the minimum length."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if self._made_find_announcement:
            context_obj, context_offset = context
            if context_obj == obj and context_offset == offset:
                msg = "DOCUMENT PRESENTER: Find result location is unchanged. Not presenting."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return False

            if self.get_only_speak_changed_lines():
                context_rect = AXText.get_range_rect(
                    context_obj, context_offset, context_offset + 1
                )
                current_rect = AXText.get_range_rect(obj, offset, offset + 1)
                if AXUtilities.rects_are_on_same_line(context_rect, current_rect):
                    msg = "DOCUMENT PRESENTER: Find result is on the same line as the last one."
                    debug.print_message(debug.LEVEL_INFO, msg, True)
                    return False

        contents = script.utilities.get_line_contents_at_offset(obj, offset)
        contents = self._restrict_find_results_to_cell(obj, contents)
        presentation_manager.get_manager().speak_contents(contents)
        script.update_braille(obj)

        results_count = script.utilities.get_find_results_count()
        if results_count:
            presentation_manager.get_manager().present_message(results_count)

        self._made_find_announcement = True
        return True

    def _force_browse_mode_for_web_app_descendant(
        self,
        script: default.Script,
        obj: Atspi.Accessible,
    ) -> bool:
        """Returns True if we should force browse mode for web-app descendant obj."""

        if not AXUtilities.is_embedded_descendant(obj):
            return False

        if AXUtilities.is_tool_tip(obj):
            return AXUtilities.is_focused(obj)

        document = AXUtilities.find_ancestor_inclusive(obj, AXUtilities.is_document)
        if document is not None and AXUtilities.is_embedded_descendant(document):
            return not self.is_focus_mode_widget(script, obj)

        return False

    def _navigation_prevents_focus_mode(
        self,
        script: default.Script,
        obj: Atspi.Accessible,
        prev_obj: Atspi.Accessible | None,
    ) -> tuple[bool | None, str]:
        """Returns (True/False, reason) if navigation state determines focus mode.

        Returns (None, '') if navigation doesn't determine the result.
        """

        if self.focus_mode_is_sticky(script.app):
            return True, "focus mode is sticky"
        if self.browse_mode_is_sticky(script.app):
            return False, "browse mode is sticky"
        if focus_manager.get_manager().in_say_all():
            return self.in_focus_mode(script.app), "SayAll preserves current mode"
        if table_navigator.get_navigator().last_input_event_was_navigation_command():
            return self.in_focus_mode(script.app), "table nav preserves current mode"

        _structural_navigator = structural_navigator.get_navigator()
        _caret_navigator = caret_navigator.get_navigator()
        caret_prevents = (
            _caret_navigator.last_command_prevents_focus_mode()
            and not AXUtilities.is_tool_tip_descendant(prev_obj, inclusive=True)
        )
        if _structural_navigator.last_command_prevents_focus_mode() or caret_prevents:
            struct_prevents = _structural_navigator.last_command_prevents_focus_mode()
            nav_type = "structural" if struct_prevents else "caret"
            return False, f"prevented by {nav_type} nav settings"

        old_doc = script.utilities.get_top_level_document_for_object(prev_obj)
        new_doc = script.utilities.get_top_level_document_for_object(obj)
        if old_doc == new_doc and not self.get_native_nav_triggers_focus_mode():
            was_struct_nav = _structural_navigator.last_input_event_was_navigation_command()
            was_caret_nav = _caret_navigator.last_input_event_was_navigation_command()
            if not (was_struct_nav or was_caret_nav):
                result = self.in_focus_mode(script.app)
                return result, "prevented by native nav settings"

        return None, ""

    def use_focus_mode(
        self,
        obj: Atspi.Accessible,
        prev_obj: Atspi.Accessible | None = None,
    ) -> bool:
        """Returns True if we should use focus mode in obj."""

        script = script_manager.get_manager().get_active_script()
        if script is None:
            msg = "DOCUMENT PRESENTER: No active script for focus mode decision."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if prev_obj and AXObject.is_dead(prev_obj):
            prev_obj = None

        nav_result, reason = self._navigation_prevents_focus_mode(script, obj, prev_obj)
        if nav_result is not None:
            prefix = "Using" if nav_result else "Not using"
            tokens = ["DOCUMENT PRESENTER:", prefix, "focus mode:", reason]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return nav_result

        do_not_toggle = AXUtilities.is_link(obj) or AXUtilities.is_radio_button(obj)
        stay_in_focus = (
            self.in_focus_mode(script.app)
            and do_not_toggle
            and input_event_manager.get_manager().last_event_was_unmodified_arrow()
        )
        if self.is_focus_mode_widget(script, obj) or stay_in_focus:
            reason = "is a focus mode widget" if not stay_in_focus else "arrowing in link/radio"
            tokens = ["DOCUMENT PRESENTER: Using focus mode:", obj, reason]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True

        was_in_app = AXUtilities.is_embedded_descendant(prev_obj)
        is_in_app = AXUtilities.is_embedded_descendant(obj)
        if is_in_app:
            if self._force_browse_mode_for_web_app_descendant(script, obj):
                tokens = ["DOCUMENT PRESENTER: Forcing browse mode for web app descendant", obj]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return False
            if not was_in_app:
                msg = "DOCUMENT PRESENTER: Using focus mode: just entered a web application"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return True
            if self.in_focus_mode(script.app):
                msg = "DOCUMENT PRESENTER: Staying in focus mode: inside a web application"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return True
            if self.user_last_changed_mode_in_document(script, obj):
                msg = "DOCUMENT PRESENTER: Not using focus mode: user changed the mode here"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return False
            msg = "DOCUMENT PRESENTER: Using focus mode: focus moved inside a web application"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        tokens = ["DOCUMENT PRESENTER: Not using focus mode for", obj, "due to lack of cause"]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return False

    def create_preferences_grid(
        self,
        title_change_callback: Callable[[str], None] | None = None,
    ) -> DocumentPreferencesGrid:
        """Returns the preferences grid for document settings."""

        # pylint: disable-next=import-outside-toplevel
        from .document_presenter_preferences_grid import DocumentPreferencesGrid

        return DocumentPreferencesGrid(self, title_change_callback)


_presenter = DocumentPresenter()


def get_presenter() -> DocumentPresenter:
    """Returns the Document Presenter."""

    return _presenter



# Orca
#
# Copyright 2011-2024 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-locals
# pylint: disable=too-many-lines

"""Manager for accessible object events."""

from __future__ import annotations

import enum
import itertools
import queue
import threading
import time
from typing import TYPE_CHECKING, Any

import gi

gi.require_version("Atspi", "2.0")
from gi.repository import Atspi, GLib

from . import (
    braille_presenter,
    dbus_service,
    debug,
    focus_manager,
    input_event,
    input_event_manager,
    orca_modifier_manager,
    script_manager,
    systemd,
)
from .ax_object import AXObject
from .ax_utilities import AXUtilities
from .ax_utilities_debugging import AXUtilitiesDebugging

if TYPE_CHECKING:
    from .scripts import default


class EventPriority(enum.IntEnum):
    """Priority levels for accessible object events."""

    IMMEDIATE = enum.auto()
    IMPORTANT = enum.auto()
    HIGH = enum.auto()
    MEDIUM_HIGH = enum.auto()
    NORMAL = enum.auto()
    LOWER = enum.auto()
    LOW = enum.auto()


class EventManager:
    """Manager for accessible object events."""

    _SKIPPABLE_SAME_TYPE_PREFIXES = (
        "document:page-changed",
        "object:active-descendant-changed",
        "object:children-changed",
        "object:property-change",
        "object:selection-changed",
        "object:state-changed",
        "object:text-caret-moved",
        "object:text-selection-changed",
        "window",
    )

    _SKIPPABLE_SIBLING_PREFIXES = ("object:state-changed:focused",)

    _SKIPPABLE_WINDOW_PREFIXES = (
        "window:activate",
        "window:deactivate",
    )

    def __init__(self) -> None:
        debug.print_message(debug.LEVEL_INFO, "EVENT MANAGER: Initializing", True)
        self._script_listener_counts: dict[str, int] = {}
        self._active: bool = False
        self._paused: bool = False
        self._counter = itertools.count()
        self._event_queue: queue.PriorityQueue[tuple[EventPriority, int, Atspi.Event]] = (
            queue.PriorityQueue(0)
        )
        self._gidle_id: int = 0
        self._gidle_lock = threading.Lock()
        self._listener: Atspi.EventListener = Atspi.EventListener.new(self._enqueue_object_event)
        self._event_history: dict[str, tuple[int | None, float]] = {}
        self._latest_event: dict[tuple[str, int], int] = {}
        dbus_service.get_remote_controller().register_decorated_module("EventManager", self)
        debug.print_message(debug.LEVEL_INFO, "Event manager initialized", True)

    def is_idle(self) -> bool:
        """Returns True if the object-event queue is empty and nothing is queued to process."""

        with self._gidle_lock:
            return self._event_queue.empty() and self._gidle_id == 0

    @dbus_service.testing_command
    def is_idle_for_testing(
        self,
        token: str = "",  # pylint: disable=unused-argument
        script: default.Script | None = None,  # pylint: disable=unused-argument
        event: input_event.InputEvent | None = None,  # pylint: disable=unused-argument
    ) -> bool:
        """Returns True if Orca has finished processing queued events (test-only)."""

        return self.is_idle()

    def activate(self) -> None:
        """Called when this event manager is activated."""

        debug.print_message(debug.LEVEL_INFO, "EVENT MANAGER: Activating", True, True)
        if self._active:
            debug.print_message(debug.LEVEL_INFO, "EVENT MANAGER: Already activated", True)
            return

        input_event_manager.get_manager().start_key_watcher()
        orca_modifier_manager.get_manager().add_grabs_for_orca_modifiers()
        self._active = True
        debug.print_message(debug.LEVEL_INFO, "EVENT MANAGER: Activated", True)

    def deactivate(self) -> None:
        """Called when this event manager is deactivated."""

        debug.print_message(debug.LEVEL_INFO, "EVENT MANAGER: Deactivating", True, True)
        if not self._active:
            debug.print_message(debug.LEVEL_INFO, "EVENT MANAGER: Already deactivated", True)
            return

        input_event_manager.get_manager().stop_key_watcher()
        self._active = False
        with self._gidle_lock:
            self._event_queue = queue.PriorityQueue(0)
            self._latest_event = {}
        self._script_listener_counts = {}
        debug.print_message(debug.LEVEL_INFO, "EVENT MANAGER: Deactivated", True)

    def pause_queuing(
        self,
        pause: bool = True,
        clear_queue: bool = False,
        reason: str = "",
    ) -> None:
        """Pauses/unpauses event queuing."""

        tokens = [
            "EVENT MANAGER: Pause queueing:",
            pause,
            ". Clear queue:",
            clear_queue,
            ".",
            reason,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        self._paused = pause
        if clear_queue:
            with self._gidle_lock:
                self._event_queue = queue.PriorityQueue(0)
                self._latest_event = {}
        input_event_manager.get_manager().pause_key_watcher(pause, reason)

    def _get_text_changed_priority(self, event: Atspi.Event) -> EventPriority:
        """Returns the priority of a text-changed event, which is higher for live regions."""

        if not event.type.startswith("object:text-changed:insert"):
            return EventPriority.MEDIUM_HIGH

        if not AXUtilities.has_live_region_role(event.source):
            return EventPriority.MEDIUM_HIGH

        live = AXObject.get_attribute(event.source, "container-live")
        if live == "assertive":
            return EventPriority.IMPORTANT
        if live == "polite":
            return EventPriority.HIGH
        return EventPriority.MEDIUM_HIGH

    def _get_priority(self, event: Atspi.Event) -> EventPriority:
        """Returns the priority associated with event."""

        event_type = event.type
        if event_type.startswith("window") or (
            event_type == "object:state-changed:active"
            and (AXUtilities.is_frame(event.source) or AXUtilities.is_dialog_or_alert(event.source))
        ):
            priority = EventPriority.IMPORTANT
        elif event_type.startswith(
            ("object:state-changed:focused", "object:active-descendant-changed"),
        ):
            priority = EventPriority.HIGH
        elif event_type.startswith("object:announcement"):
            if event.detail1 == Atspi.Live.ASSERTIVE:
                priority = EventPriority.IMPORTANT
            elif event.detail1 == Atspi.Live.POLITE:
                priority = EventPriority.HIGH
            else:
                priority = EventPriority.NORMAL
        elif event_type.startswith("object:state-changed:invalid-entry"):
            # Setting this to lower ensures we present the state and/or text changes that triggered
            # the invalid state prior to presenting the invalid state.
            priority = EventPriority.LOWER
        elif event_type.startswith("object:children-changed"):
            priority = EventPriority.LOW
        elif event_type.startswith("object:text-changed"):
            priority = self._get_text_changed_priority(event)
        elif event_type.startswith("object:property-change:accessible-description"):
            priority = EventPriority.LOWER
        else:
            priority = EventPriority.NORMAL

        tokens = ["EVENT MANAGER:", event, "has priority level:", priority.name]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return priority

    # pylint: disable-next=too-many-return-statements
    def _is_obsoleted_by(self, event: Atspi.Event, counter: int = -1) -> Atspi.Event | None:
        """Returns the event which renders this one no longer worthy of being processed."""

        if event.type.startswith(EventManager._SKIPPABLE_SAME_TYPE_PREFIXES):
            key = (event.type, hash(event.source))
            latest = self._latest_event.get(key, -1)
            if latest > counter >= 0:
                tokens = [
                    "EVENT MANAGER:",
                    event,
                    "(#",
                    counter,
                    ") obsoleted: a newer",
                    event.type,
                    "for same source is queued (#",
                    latest,
                    ")",
                ]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return event

        event_is_skippable_sibling = event.type.startswith(EventManager._SKIPPABLE_SIBLING_PREFIXES)
        event_is_window = event.type.startswith(EventManager._SKIPPABLE_WINDOW_PREFIXES)

        if not event_is_skippable_sibling and not event_is_window:
            return None

        def is_same(x):
            return (
                x.type == event.type
                and x.source == event.source
                and x.detail1 == event.detail1
                and x.detail2 == event.detail2
                and x.any_data == event.any_data
            )

        def obsoletes_if_same_type_in_sibling(x):
            if not event_is_skippable_sibling:
                return False
            if (
                x.type != event.type
                or x.detail1 != event.detail1
                or x.detail2 != event.detail2
                or x.any_data != event.any_data
            ):
                return False
            return AXObject.get_parent(x.source) == AXObject.get_parent(event.source)

        def obsoletes_window_event(x):
            if not event_is_window:
                return False
            if x.source != event.source:
                return False
            return x.type.startswith(EventManager._SKIPPABLE_WINDOW_PREFIXES)

        with self._event_queue.mutex:
            try:
                events = list(reversed(self._event_queue.queue))
            except queue.Empty as error:
                tokens = ["EVENT MANAGER: Exception in _is_obsoleted_by:", error]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                events = []

        for _priority, _counter, e in events:
            if e == event:
                return None
            if is_same(e):
                tokens = ["EVENT MANAGER:", event, "obsoleted by", e, "more recent duplicate"]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return e
            if obsoletes_if_same_type_in_sibling(e):
                tokens = [
                    "EVENT MANAGER:",
                    event,
                    "obsoleted by",
                    e,
                    "more recent event of same type from sibling",
                ]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return e
            if obsoletes_window_event(e):
                tokens = [
                    "EVENT MANAGER:",
                    event,
                    "obsoleted by",
                    e,
                    "more recent window (de)activation event",
                ]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return e

        return None

    def _ignore_by_role(self, event: Atspi.Event, role: Atspi.Role) -> bool | None:
        """Returns True/False if the source role determines ignore, or None if inconclusive."""

        event_type = event.type

        # gnome-shell fires "focused" events spuriously after the Alt+Tab switcher
        # is used and something else has claimed focus.
        if AXUtilities.is_window(event.source, role) and "focused" in event_type:
            tokens = ["EVENT MANAGER: Ignoring", event_type, "based on type and role"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True

        if AXUtilities.is_frame(event.source, role):
            app = AXUtilities.get_application(event.source)
            ignore = AXUtilities.is_mutter_x11_frames(app)
            prefix = "Ignoring" if ignore else "Not ignoring"
            reason = "application" if ignore else "role"
            tokens = ["EVENT MANAGER:", prefix, event_type, "based on", reason]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return ignore

        # Events from the text role are typically something we want to handle.
        # One exception is a huge text insertion.
        if AXUtilities.is_text(event.source, role):
            if event_type.startswith("object:text-changed:insert") and event.detail2 > 5000:
                tokens = ["EVENT_MANAGER: Ignoring", event_type, "due to size of inserted text"]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return True
            if not event_type.startswith("object:text-caret-moved"):
                tokens = ["EVENT_MANAGER: Not ignoring", event_type, "due to role"]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return False

        if AXUtilities.is_notification(event.source, role) or AXUtilities.is_alert(
            event.source, role
        ):
            tokens = ["EVENT_MANAGER: Not ignoring", event_type, "due to role"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        return None

    def _ignore_by_focus_state(
        self,
        event: Atspi.Event,
        focus: Atspi.Accessible | None,
        role: Atspi.Role,
    ) -> bool | None:
        """Returns False if focus/state means we should not ignore, or None if inconclusive."""

        event_type = event.type
        if focus in (event.source, event.any_data):
            reason = "source" if focus == event.source else "any_data"
            tokens = [
                "EVENT_MANAGER: Not ignoring",
                event_type,
                "due to",
                reason,
                "being locus of focus",
            ]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        source_states = AXObject.get_state_set(event.source)
        if AXUtilities.is_selected(event.source, source_states):
            tokens = ["EVENT_MANAGER: Not ignoring", event_type, "due to source being selected"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        # We see an unbelievable number of active-descendant-changed and selection changed
        # from Caja when the user navigates from one giant folder to another. We need the
        # spam filtering below to catch this bad behavior coming from a focused object, so
        # only return early here if the focused object doesn't manage descendants, or the
        # event is not a focus claim.
        if AXUtilities.is_focused(event.source, source_states):
            if not AXUtilities.manages_descendants(event.source, source_states) or (
                event_type.startswith("object:state-changed:focused") and event.detail1
            ):
                tokens = ["EVENT_MANAGER: Not ignoring", event_type, "due to source being focused"]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return False

        if event_type.startswith("object:text-changed:insert") and AXUtilities.has_live_region_role(
            event.source, role
        ):
            live = AXObject.get_attribute(event.source, "live")
            if live and live != "off":
                tokens = [
                    "EVENT_MANAGER: Not ignoring",
                    event_type,
                    "due to source being live region",
                ]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return False

        return None

    def _ignore_by_spam_filter(self, event: Atspi.Event) -> bool | None:
        """Returns True if the event is spam, or None if inconclusive."""

        event_type = event.type
        last_app, last_time = self._event_history.get(event_type, (None, 0))
        app = AXUtilities.get_application(event.source)
        ignore = last_app == hash(app) and time.time() - last_time < 0.1
        self._event_history[event_type] = hash(app), time.time()
        if ignore:
            tokens = [
                "EVENT_MANAGER: Ignoring",
                event_type,
                "due to multiple instances in short time",
            ]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True

        if AXUtilities.is_mutter_x11_frames(app):
            tokens = ["EVENT MANAGER: Ignoring", event_type, "based on application"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True

        return None

    def _ignore_children_changed(
        self,
        event: Atspi.Event,
        focus: Atspi.Accessible | None,
    ) -> bool | None:
        """Returns True/False for children-changed events, or None if not applicable."""

        event_type = event.type
        if not event_type.startswith("object:children-changed"):
            return None

        if "remove" in event_type:
            if (focus and AXObject.is_dead(focus)) or event.source == AXUtilities.get_desktop():
                return False

        child = event.any_data
        if child is None or AXObject.is_dead(child):
            tokens = ["EVENT_MANAGER: Ignoring", event_type, "due to null/dead event.any_data"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True
        if AXUtilities.is_menu_related(child) or AXUtilities.is_image(child):
            tokens = ["EVENT_MANAGER: Ignoring", event_type, "due to role of event.any_data"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True

        script = script_manager.get_manager().get_active_script()
        if script is None or script.app != AXUtilities.get_application(event.source):
            reason = (
                "there is no active script" if script is None else "event is not from active app"
            )
            tokens = ["EVENT MANAGER: Ignoring", event_type, "because", reason]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True

        return None

    @staticmethod
    def _ignore_property_change(event: Atspi.Event, role: Atspi.Role) -> bool | None:
        """Returns True/False for property-change events, or None if not applicable."""

        event_type = event.type
        if not event_type.startswith("object:property-change"):
            return None

        if "name" in event_type:
            ignore_name_roles = [
                Atspi.Role.CANVAS,
                Atspi.Role.CHECK_BOX,
                Atspi.Role.ICON,
                Atspi.Role.IMAGE,
                Atspi.Role.LIST,
                Atspi.Role.LIST_ITEM,
                Atspi.Role.MENU,
                Atspi.Role.MENU_ITEM,
                Atspi.Role.PANEL,
                Atspi.Role.RADIO_BUTTON,
                Atspi.Role.SECTION,
                Atspi.Role.TABLE_ROW,
                Atspi.Role.TABLE_CELL,
                Atspi.Role.TREE_ITEM,
            ]
            if role in ignore_name_roles:
                tokens = ["EVENT MANAGER: Ignoring", event_type, "due to role of unfocused source"]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return True
            return False
        if "value" in event_type:
            if role in [Atspi.Role.SPLIT_PANE, Atspi.Role.SCROLL_BAR]:
                tokens = ["EVENT MANAGER: Ignoring", event_type, "due to role of unfocused source"]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return True
            return False

        return None

    @staticmethod
    def _ignore_state_changed(event: Atspi.Event, role: Atspi.Role) -> bool | None:
        """Returns True/False for state-changed events, or None if not applicable."""

        event_type = event.type
        if not event_type.startswith("object:state-changed"):
            return None

        if event_type.endswith("system"):
            system_ignore_roles = [
                Atspi.Role.TABLE,
                Atspi.Role.TABLE_CELL,
                Atspi.Role.TABLE_ROW,
                Atspi.Role.TREE,
                Atspi.Role.TREE_ITEM,
                Atspi.Role.TREE_TABLE,
            ]
            if role in system_ignore_roles:
                tokens = ["EVENT MANAGER: Ignoring", event_type, "based on role"]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return True

        return EventManager._ignore_state_changed_subtype(event, event_type, role)

    @staticmethod
    def _ignore_state_changed_subtype(
        event: Atspi.Event,
        event_type: str,
        role: int,
    ) -> bool | None:
        """Returns True/False for state-changed subtypes, or None if not applicable."""

        showing_roles = [
            Atspi.Role.ALERT,
            Atspi.Role.ANIMATION,
            Atspi.Role.DIALOG,
            Atspi.Role.INFO_BAR,
            Atspi.Role.MENU,
            Atspi.Role.NOTIFICATION,
            Atspi.Role.STATUS_BAR,
            Atspi.Role.TOOL_TIP,
        ]

        ignore: bool | None = None
        reason = ""
        if "checked" in event_type:
            ignore = not AXUtilities.is_showing(event.source)
            reason = "unfocused, non-showing source"
        elif "selected" in event_type:
            ignore = not event.detail1 and role == Atspi.Role.BUTTON
            reason = "role of source and detail1"
        elif "sensitive" in event_type:
            ignore = role not in [Atspi.Role.TEXT, Atspi.Role.ENTRY]
            reason = "role of unfocused source"
        elif "showing" in event_type:
            ignore = role not in showing_roles
            reason = "role"

        if ignore is None:
            return None

        if ignore:
            tokens = ["EVENT MANAGER: Ignoring", event_type, "due to", reason]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return ignore

    @staticmethod
    def _ignore_active_descendant_or_selection(event: Atspi.Event) -> bool | None:
        """Returns True/False for active-descendant and selection events."""

        event_type = event.type
        if event_type.startswith("object:active-descendant-changed"):
            child = event.any_data
            if child is None or AXUtilities.is_invalid_role(child):
                tokens = [
                    "EVENT_MANAGER: Ignoring",
                    event_type,
                    "due to null/invalid event.any_data",
                ]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return True
            return False

        if event_type.startswith("object:selection-changed"):
            if AXObject.is_dead(event.source, AXUtilities.get_application(event.source)):
                tokens = ["EVENT MANAGER: Ignoring", event_type, "from dead source"]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return True
            return False

        return None

    @staticmethod
    def _ignore_text_events(
        event: Atspi.Event,
        focus: Atspi.Accessible | None,
        role: Atspi.Role,
    ) -> bool | None:
        """Returns True/False for text caret-moved and text-changed events."""

        event_type = event.type
        if event_type.startswith("object:text-caret-moved"):
            if role == Atspi.Role.LABEL:
                tokens = ["EVENT MANAGER: Ignoring", event_type, "due to role of unfocused source"]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return True
            return False

        if event_type.startswith("object:text-changed"):
            if "insert" in event_type and event.detail2 > 1000:
                tokens = ["EVENT MANAGER: Ignoring", event_type, "due to inserted text size"]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return True
            if event_type.endswith("system") and AXUtilities.is_selectable(focus):
                tokens = ["EVENT MANAGER: Ignoring because", event_type, "is suspected spam"]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return True
            return False

        return None

    def _ignore(self, event: Atspi.Event) -> bool:
        """Returns True if this event should be ignored."""

        debug.print_message(debug.LEVEL_INFO, "")
        tokens = ["EVENT MANAGER:", event]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if event.type.startswith(("window", "mouse:button")):
            return False

        if not self._active or self._paused:
            msg = "EVENT MANAGER: Ignoring because manager is not active or queueing is paused"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        focus = focus_manager.get_manager().get_locus_of_focus()
        role = AXObject.get_role(event.source)
        for check in (
            lambda: self._ignore_by_role(event, role),
            lambda: self._ignore_by_focus_state(event, focus, role),
            lambda: self._ignore_property_change(event, role),
            lambda: self._ignore_by_spam_filter(event),
            lambda: self._ignore_active_descendant_or_selection(event),
            lambda: self._ignore_children_changed(event, focus),
            lambda: self._ignore_state_changed(event, role),
            lambda: self._ignore_text_events(event, focus, role),
        ):
            result = check()
            if result is not None:
                return result

        return False

    def _queue_println(
        self,
        event: input_event.InputEvent | Atspi.Event,
        is_enqueue: bool = True,
    ) -> None:
        """Convenience method to output queue-related debugging info."""

        if debug.debugLevel > debug.LEVEL_INFO:
            return

        tokens = []
        if isinstance(event, input_event.KeyboardEvent):
            tokens.extend([event.keyval_name, event.hw_code])
        elif isinstance(event, input_event.BrailleEvent):
            tokens.append(event.event)
        else:
            tokens.append(event)

        if is_enqueue:
            tokens[0:0] = ["EVENT MANAGER: Queueing"]
        else:
            tokens[0:0] = ["EVENT MANAGER: Dequeueing"]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

    def _enqueue_object_event(self, e: Atspi.Event) -> None:
        """Callback for Atspi object events."""

        # If we are enqueuing events, we're not dead and should not be killed
        # and restarted by systemd.
        if self._event_queue.qsize() > 75 and systemd.get_manager().is_systemd_managed():
            systemd.get_manager().notify_alive("Event queue size > 75")

        if self._ignore(e):
            return

        self._queue_println(e)
        app = AXUtilities.get_application(e.source)
        tokens = ["EVENT MANAGER: App for event source is", app]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if AXObject.check_hung(e.source, app):
            tokens = ["EVENT MANAGER: Dropping", e, "from hung source or app"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return

        script = script_manager.get_manager().get_script(app, e.source)
        script.record_queued_event(e)

        priority = self._get_priority(e)
        with self._gidle_lock:
            counter = next(self._counter)
            self._event_queue.put((priority, counter, e))
            if e.type.startswith(EventManager._SKIPPABLE_SAME_TYPE_PREFIXES):
                self._latest_event[(e.type, hash(e.source))] = counter
            if not self._gidle_id:
                self._gidle_id = GLib.idle_add(self._dequeue_object_event)
        tokens = ["EVENT MANAGER: Queued", e, "priority:", priority.name, ", counter:", counter]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

    def _on_no_focus(self) -> bool:
        if not focus_manager.get_manager().focus_and_window_are_unknown():
            return False

        if script_manager.get_manager().get_active_script() is None:
            default_script = script_manager.get_manager().get_default_script()
            script_manager.get_manager().set_active_script(default_script, "No focus")
            braille_presenter.get_presenter().disable_braille()

        return False

    def _dequeue_object_event(self) -> bool:
        """Handles all object events destined for scripts."""

        rerun = True
        try:
            priority, counter, event = self._event_queue.get_nowait()
            self._queue_println(event, is_enqueue=False)
            tokens = [
                "EVENT MANAGER: Dequeued",
                event,
                "priority:",
                priority.name,
                ", counter:",
                counter,
            ]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            start_time = time.time()
            tokens = [
                "\nvvvvv START ",
                priority.name,
                "-PRIORITY OBJECT EVENT",
                event.type.upper(),
                "(queue size:",
                self._event_queue.qsize(),
                ") vvvvv",
            ]
            debug.print_tokens(debug.LEVEL_INFO, tokens, False)
            self._process_object_event(event, counter)
            tokens = [
                "TOTAL PROCESSING TIME:",
                round(time.time() - start_time, 4),
                "\n^^^^^ FINISHED ",
                priority.name,
                "-PRIORITY OBJECT EVENT",
                event.type.upper(),
                "^^^^^\n",
            ]
            debug.print_tokens(debug.LEVEL_INFO, tokens, False)
            with self._gidle_lock:
                if self._event_queue.empty():
                    GLib.timeout_add(2500, self._on_no_focus)
                    self._gidle_id = 0
                    rerun = False  # destroy and don't call again
        except queue.Empty:
            msg = "EVENT MANAGER: Attempted dequeue, but the event queue is empty"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            with self._gidle_lock:
                if self._event_queue.empty():
                    self._gidle_id = 0
                    rerun = False  # destroy and don't call again
        except Exception:  # pylint: disable=broad-except
            with self._gidle_lock:
                self._gidle_id = GLib.idle_add(self._dequeue_object_event)
            raise

        return rerun

    def register_listener(self, event_type: str) -> None:
        """Tells this module to listen for the given event type.

        Arguments:
        - event_type: the event type.
        """

        tokens = ["EVENT MANAGER: registering listener for:", event_type]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if event_type in self._script_listener_counts:
            self._script_listener_counts[event_type] += 1
        else:
            self._listener.register(event_type)
            self._script_listener_counts[event_type] = 1

    def deregister_listener(self, event_type: str) -> None:
        """Tells this module to stop listening for the given event type.

        Arguments:
        - event_type: the event type.
        """

        tokens = ["EVENT MANAGER: deregistering listener for:", event_type]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if event_type not in self._script_listener_counts:
            return

        self._script_listener_counts[event_type] -= 1
        if self._script_listener_counts[event_type] == 0:
            try:
                self._listener.deregister(event_type)
            except GLib.GError as error:
                tokens = [
                    "EVENT MANAGER: Exception deregistering listener for",
                    event_type,
                    ":",
                    error,
                ]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            del self._script_listener_counts[event_type]

    def register_script_listeners(self, script: default.Script) -> None:
        """Tells the event manager to start listening for all the event types
        of interest to the script.

        Arguments:
        - script: the script.
        """

        tokens = ["EVENT MANAGER: Registering listeners for:", script]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        for event_type in script.listeners:
            self.register_listener(event_type)

    def deregister_script_listeners(self, script: default.Script) -> None:
        """Tells the event manager to stop listening for all the event types
        of interest to the script.

        Arguments:
        - script: the script.
        """

        tokens = ["EVENT MANAGER: De-registering listeners for:", script]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        for event_type in script.listeners:
            self.deregister_listener(event_type)

    def _get_script_for_event(
        self,
        event: Atspi.Event,
        active_script: default.Script | None = None,
    ) -> default.Script | None:
        """Returns the script associated with event."""

        if event.source == focus_manager.get_manager().get_locus_of_focus():
            script = active_script or script_manager.get_manager().get_active_script()
            tokens = ["EVENT MANAGER: Script for event from locus of focus is", script]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return script

        if event.type.startswith("mouse:"):
            mouse_event = input_event.MouseButtonEvent(event)
            script = script_manager.get_manager().get_script(mouse_event.app, mouse_event.window)
            tokens = ["EVENT MANAGER: Script for event is", script]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return script

        script = None
        app = AXUtilities.get_application(event.source)
        if AXUtilities.is_defunct(app):
            tokens = ["EVENT MANAGER:", app, "is defunct. Cannot get script for event."]
            debug.print_tokens(debug.LEVEL_WARNING, tokens, True)
            return None

        tokens = ["EVENT MANAGER: Getting script for event for", app, event.source]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        script = script_manager.get_manager().get_script(app, event.source)
        tokens = ["EVENT MANAGER: Script for event is", script]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return script

    def _is_activatable_event(
        self,
        event: Atspi.Event,
        script: default.Script | None = None,
    ) -> tuple[bool, str]:
        """Determines if event should cause us to change the active script."""

        if not event.source:
            return False, "event.source? What event.source??"

        if not script:
            script = self._get_script_for_event(event)
            if not script:
                return False, "There is no script for this event."

        app = AXUtilities.get_application(event.source)
        if app and not AXUtilities.is_application_in_desktop(app):
            return False, "The application is unknown to AT-SPI2"

        if not script.is_activatable_event(event):
            return False, "The script says not to activate for this event."

        if script.force_script_activation(event):
            return True, "The script insists it should be activated for this event."

        return self._is_activatable_by_event_type(event)

    @staticmethod
    def _is_activatable_by_event_type(event: Atspi.Event) -> tuple[bool, str]:
        """Determines if event type makes this an activatable event."""

        event_type = event.type

        window_activation = event_type.startswith("window:activate") or (
            event_type.startswith("object:state-changed:active")
            and event.detail1
            and AXUtilities.is_frame(event.source)
        )
        if window_activation:
            is_new = event.source != focus_manager.get_manager().get_active_window()
            reason = (
                "Window activation" if is_new else "Window activation for already-active window"
            )
            return is_new, reason

        if event_type.startswith("object:state-changed:focused") and event.detail1:
            return True, "Event source claimed focus."

        if (
            event_type.startswith("object:state-changed:selected")
            and event.detail1
            and AXUtilities.is_menu(event.source)
            and AXUtilities.is_focusable(event.source)
        ):
            return True, "Selection change in focused menu"

        # This condition appears with gnome-screensaver-dialog. See bug 530368.
        if (
            event_type.startswith("object:state-changed:showing")
            and AXUtilities.is_panel(event.source)
            and AXUtilities.is_modal(event.source)
        ):
            return True, "Modal panel is showing."

        return False, "No reason found to activate a different script."

    def _event_source_is_dead(self, event: Atspi.Event) -> bool:
        if AXObject.is_dead(event.source, AXUtilities.get_application(event.source)):
            tokens = ["EVENT MANAGER: source of", event.type, "is dead"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True

        return False

    def _should_process_event(
        self,
        event: Atspi.Event,
        event_script: default.Script,
        active_script: default.Script | None,
    ) -> bool:
        """Returns True if this event should be processed."""

        if event_script == active_script:
            tokens = ["EVENT MANAGER: Processing", event.type, ": script for event is active"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True

        if event_script.present_if_inactive:
            tokens = [
                "EVENT MANAGER: Processing",
                event.type,
                ": script handles events when inactive",
            ]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True

        if "accessible-value" in event.type and AXUtilities.is_progress_bar(event.source):
            tokens = ["EVENT MANAGER: Processing", event.type, ": source is progress bar"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True

        tokens = ["EVENT MANAGER: Not processing", event.type, "due to lack of reason"]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return False

    @staticmethod
    def _handle_early_event_processing(event: Atspi.Event) -> bool:
        """Handles early event processing. Returns True if the event was fully handled."""

        script_mgr = script_manager.get_manager()
        focus_mgr = focus_manager.get_manager()

        event_type = event.type
        if (
            event_type.startswith("object:children-changed:remove")
            and event.source == AXUtilities.get_desktop()
        ):
            script_mgr.reclaim_scripts()
            return True

        app = AXUtilities.get_application(event.source)
        if AXObject.is_dead(event.source, app) or AXUtilities.is_defunct(event.source):
            if (
                event_type.startswith("window:activate")
                and AXObject.object_is_known_dead(event.source)
                and AXObject.revalidate_if_known_dead(event.source)
            ):
                app = AXUtilities.get_application(event.source)
                if not AXObject.is_dead(event.source, app) and not AXUtilities.is_defunct(
                    event.source
                ):
                    return False

            tokens = ["EVENT MANAGER: Ignoring defunct object:", event.source]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)

            if event_type.startswith("window:de") and focus_mgr.get_active_window() == event.source:
                focus_mgr.clear_state("Active window is dead or defunct")
                script_mgr.set_active_script(None, "Active window is dead or defunct")
            return True

        if event_type.startswith("window:") and event_type.endswith("destroy"):
            script_mgr.reclaim_scripts()

        if AXUtilities.is_iconified(event.source):
            tokens = ["EVENT MANAGER: Ignoring iconified object:", event.source]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True

        return False

    @staticmethod
    def _find_listener(script: default.Script, event_type: str):
        """Returns the listener for event_type, or None."""

        listener = script.listeners.get(event_type)
        if listener is not None:
            return listener

        # The listener can be None if the event type has a suffix such as "system".
        for key, value in script.listeners.items():
            if event_type.startswith(key):
                return value

        return None

    def _process_object_event(self, event: Atspi.Event, counter: int = -1) -> None:
        """Handles all object events destined for scripts."""

        if self._is_obsoleted_by(event, counter) or self._handle_early_event_processing(event):
            return

        if debug.debugLevel <= debug.LEVEL_INFO:
            # The details are still one string. They become tokens when the debug value tree
            # can carry them.
            tokens: list[Any] = [AXUtilitiesDebugging.object_event_details_as_string(event)]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        script_mgr = script_manager.get_manager()
        active_script = script_mgr.get_active_script()
        script = self._get_script_for_event(event, active_script)
        if not script:
            msg = "ERROR: Could not get script for event"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return

        if script != active_script:
            set_new_active_script, reason = self._is_activatable_event(event, script)
            tokens = [
                "EVENT MANAGER: Change active script:",
                set_new_active_script,
                "(",
                reason,
                ")",
            ]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)

            if set_new_active_script:
                script_mgr.set_active_script(script, reason)
                active_script = script

        if not self._should_process_event(event, script, active_script):
            return

        listener = self._find_listener(script, event.type)
        if listener is None:
            tokens = ["EVENT MANAGER: No listener for event type", event.type]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return

        listener(event)


_manager: EventManager = EventManager()


def get_manager() -> EventManager:
    """Returns the Event Manager singleton."""
    return _manager



# Orca
#
# Copyright 2024 Igalia, S.L.
# Copyright 2024 GNOME Foundation Inc.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Wrapper for the Atspi.Document interface."""

from __future__ import annotations

import gi

gi.require_version("Atspi", "2.0")
from gi.repository import Atspi, GLib

from . import debug
from .ax_object import AXObject


class AXDocument:
    """Wrapper for the Atspi.Document interface."""

    @staticmethod
    def get_text_selections(
        document: Atspi.Accessible,
    ) -> tuple[bool, list[Atspi.TextSelection]]:
        """Returns whether the getter succeeded and the document's text selections."""

        if not AXObject.supports_document(document):
            return False, []

        atspi_version = Atspi.get_version()  # pylint: disable=no-value-for-parameter
        getter_is_safe = (
            atspi_version >= (2, 61, 2)
            or (atspi_version[:2] == (2, 60) and atspi_version[2] >= 7)
            or (atspi_version[:2] == (2, 58) and atspi_version[2] >= 9)
        )
        if getter_is_safe:
            try:
                result = Atspi.Document.get_text_selections(document)
            except GLib.GError as error:
                tokens = ["AXDocument: Exception in get_text_selections:", error]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return False, []

            selections = list(result or [])
            tokens = ["AXDocument:", document, "reports", len(selections), "text selection(s)."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True, selections

        # Older versions can return dangling accessible pointers:
        # https://gitlab.gnome.org/GNOME/at-spi2-core/-/work_items/243
        version_string = ".".join(str(part) for part in atspi_version)
        tokens = [
            "AXDocument: Not getting text selections due to at-spi2-core issue 243. Version:",
            version_string,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return False, []

    @staticmethod
    def set_text_selection(
        document: Atspi.Accessible,
        start_object: Atspi.Accessible,
        start_offset: int,
        end_object: Atspi.Accessible,
        end_offset: int,
        start_is_active: bool,
    ) -> bool:
        """Sets a single document text selection."""

        if not AXObject.supports_document(document):
            return False

        selection = Atspi.TextSelection()
        selection.start_object = start_object
        selection.start_offset = start_offset
        selection.end_object = end_object
        selection.end_offset = end_offset
        selection.start_is_active = start_is_active

        try:
            result = Atspi.Document.set_text_selections(document, [selection])
        except GLib.GError as error:
            tokens = ["AXDocument: Exception in set_text_selection:", error]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        tokens = [
            "AXDocument: Set text selection in",
            document,
            "from",
            start_object,
            start_offset,
            "to",
            end_object,
            end_offset,
            "with start active:",
            start_is_active,
            ". Result:",
            result,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return result

    @staticmethod
    def get_page_count(document: Atspi.Accessible) -> int:
        """Returns the page count of document."""

        if not AXObject.supports_document(document):
            return 0

        try:
            count = Atspi.Document.get_page_count(document)
        except GLib.GError as error:
            tokens = ["AXDocument: Exception in get_page_count:", error]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return 0

        tokens = ["AXDocument: Page count of", document, "is", count]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return count

    @staticmethod
    def get_locale(document: Atspi.Accessible) -> str:
        """Returns the locale of document."""

        if not AXObject.supports_document(document):
            return ""

        try:
            result = Atspi.Document.get_locale(document)
        except GLib.GError as error:
            tokens = ["AXDocument: Exception in get_locale:", error]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return ""

        if result is None:
            tokens = ["AXDocument: get_locale failed for", document]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return ""

        tokens = ["AXDocument: Locale of", document, "is '", result, "'"]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return result

    @staticmethod
    def get_attributes_dict(document: Atspi.Accessible) -> dict[str, str]:
        """Returns a dict with the document-attributes of document."""

        if not AXObject.supports_document(document):
            return {}

        try:
            result = Atspi.Document.get_document_attributes(document)
        except GLib.GError as error:
            tokens = ["AXDocument: Exception in get_attributes_dict:", error]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return {}

        tokens = ["AXDocument: Attributes of", document, "are:", result]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return result or {}



# Orca
#
# Copyright 2004-2009 Sun Microsystems Inc.
# Copyright 2011-2026 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-lines
# pylint: disable=too-many-arguments
# pylint: disable=too-many-locals
# pylint: disable=too-many-positional-arguments
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-return-statements

"""The default Script for presenting information to the user."""

from __future__ import annotations

import re
from typing import TYPE_CHECKING, Any

from orca import (
    action_presenter,
    braille,
    braille_presenter,
    bypass_mode_manager,
    caret_navigator,
    chat_presenter,
    clipboard,
    command_manager,
    debug,
    debugging_tools_manager,
    document_presenter,
    event_manager,
    extension_loader,
    flat_review_finder,
    flat_review_presenter,
    focus_manager,
    guilabels,
    input_event_manager,
    learn_mode_presenter,
    live_region_presenter,
    math_navigator,
    messages,
    mouse_presenter,
    notification_presenter,
    object_navigator,
    presentation_manager,
    profile_manager,
    say_all_presenter,
    screen_reader_manager,
    script,
    script_manager,
    sleep_mode_manager,
    speech_manager,
    speech_presenter,
    spellcheck_presenter,
    structural_navigator,
    system_information_presenter,
    table_navigator,
    text_selection_manager,
    text_selection_presenter,
    typing_echo_presenter,
    where_am_i_presenter,
)
from orca.ax_object import AXObject
from orca.ax_selection import AXSelection
from orca.ax_text import AXText
from orca.ax_utilities import AXUtilities
from orca.ax_utilities_event import TextEventReason
from orca.ax_utilities_text import CaretSetReason, TextUnit
from orca.generator import PresentationReason

if TYPE_CHECKING:
    import gi

    gi.require_version("Atspi", "2.0")
    from gi.repository import Atspi


class Script(script.Script):
    """The default Script for presenting information to the user."""

    _commands_initialized: bool = False

    def _register_builtin_extensions(self) -> None:
        """Registers built-in extensions with the extension loader."""

        loader = extension_loader.get_loader()
        builtins = [
            (braille_presenter.get_presenter, guilabels.BRAILLE),
            (notification_presenter.get_presenter, guilabels.KB_GROUP_NOTIFICATIONS),
            (clipboard.get_presenter, guilabels.KB_GROUP_CLIPBOARD),
            (say_all_presenter.get_presenter, guilabels.GENERAL_SAY_ALL),
            (typing_echo_presenter.get_presenter, guilabels.ECHO),
            (speech_manager.get_manager, guilabels.KB_GROUP_SPEECH_VERBOSITY),
            (speech_presenter.get_presenter, guilabels.KB_GROUP_SPEECH_VERBOSITY),
            (bypass_mode_manager.get_manager, guilabels.KB_GROUP_BYPASS_MODE),
            (sleep_mode_manager.get_manager, guilabels.KB_GROUP_SLEEP_MODE),
            (system_information_presenter.get_presenter, guilabels.KB_GROUP_SYSTEM_INFORMATION),
            (object_navigator.get_navigator, guilabels.KB_GROUP_OBJECT_NAVIGATION),
            (caret_navigator.get_navigator, guilabels.KB_GROUP_CARET_NAVIGATION),
            (structural_navigator.get_navigator, guilabels.KB_GROUP_STRUCTURAL_NAVIGATION),
            (table_navigator.get_navigator, guilabels.KB_GROUP_TABLE_NAVIGATION),
            (math_navigator.get_navigator, guilabels.KB_GROUP_MATH_NAVIGATION),
            (document_presenter.get_presenter, guilabels.KB_GROUP_DOCUMENTS),
            (live_region_presenter.get_presenter, guilabels.KB_GROUP_LIVE_REGIONS),
            (learn_mode_presenter.get_presenter, guilabels.KB_GROUP_LEARN_MODE),
            (mouse_presenter.get_presenter, guilabels.KB_GROUP_MOUSE),
            (action_presenter.get_presenter, guilabels.KB_GROUP_ACTIONS),
            (flat_review_presenter.get_presenter, guilabels.KB_GROUP_FLAT_REVIEW),
            (flat_review_finder.get_finder, guilabels.KB_GROUP_FIND),
            (where_am_i_presenter.get_presenter, guilabels.KB_GROUP_WHERE_AM_I),
            (debugging_tools_manager.get_manager, guilabels.KB_GROUP_DEBUGGING_TOOLS),
            (chat_presenter.get_presenter, guilabels.KB_GROUP_CHAT),
            (profile_manager.get_manager, guilabels.GENERAL_PROFILES),
            (screen_reader_manager.get_manager, guilabels.KB_GROUP_SCREEN_READER_MANAGEMENT),
        ]
        for getter, group_label in builtins:
            loader.register_builtin(getter, group_label)

    def set_up_commands(self) -> None:
        """Sets up commands with CommandManager."""

        tokens = ["DEFAULT: Setting up commands for", self.app]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if Script._commands_initialized:
            msg = "DEFAULT: Commands already initialized."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return

        Script._commands_initialized = True

        manager = command_manager.get_manager()
        manager.set_up_commands()

        self._register_builtin_extensions()
        extension_loader.get_loader().set_up_all_commands()

        all_braille_keys: set[int] = set()
        for cmd in manager.get_all_braille_commands():
            all_braille_keys.update(cmd.get_braille_bindings())
        braille.setup_key_ranges(all_braille_keys)

        cmd_count = len(command_manager.get_manager().get_all_keyboard_commands())
        tokens = ["DEFAULT: Commands set up:", cmd_count, "keyboard commands"]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

    def register_event_listeners(self) -> None:
        """Registers for listeners needed by this script."""

        event_manager.get_manager().register_script_listeners(self)

    def deregister_event_listeners(self) -> None:
        """De-registers the listeners needed by this script."""

        event_manager.get_manager().deregister_script_listeners(self)

    def _get_queued_event(
        self,
        event_type: str,
        detail1: int | None = None,
        detail2: int | None = None,
        any_data=None,
    ) -> Atspi.Event | None:
        cached_event = self.get_queued_event(event_type)
        if not cached_event:
            tokens: list[Any] = ["SCRIPT: No queued event of type", event_type]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return None

        if detail1 is not None and detail1 != cached_event.detail1:
            tokens = [
                "SCRIPT: Queued event's detail1 (",
                cached_event.detail1,
                ") doesn't match",
                detail1,
            ]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return None

        if detail2 is not None and detail2 != cached_event.detail2:
            tokens = [
                "SCRIPT: Queued event's detail2 (",
                cached_event.detail2,
                ") doesn't match",
                detail2,
            ]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return None

        if any_data is not None and any_data != cached_event.any_data:
            tokens = [
                "SCRIPT: Queued event's any_data (",
                cached_event.any_data,
                ") doesn't match",
                any_data,
            ]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return None

        tokens = ["SCRIPT: Found matching queued event:", cached_event]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return cached_event

    def locus_of_focus_changed(self, event, old_focus, new_focus):
        """Called when the visual object with focus changes."""

        presentation_manager.get_manager().present_command_announcement()

        if not new_focus:
            return True

        if AXUtilities.is_defunct(new_focus):
            return True

        is_name_change = event and event.type.endswith("accessible-name")
        if old_focus == new_focus and not is_name_change:
            msg = "DEFAULT: old focus == new focus"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if self.run_find_command_on:
            if self.run_find_command_on == new_focus:
                self.run_find_command_on = None
                flat_review_finder.get_finder().find(self)
            return True

        if learn_mode_presenter.get_presenter().is_active():
            learn_mode_presenter.get_presenter().quit()

        document_presenter.get_presenter().update_mode_if_needed(self, old_focus, new_focus)

        active_window = self.utilities.top_level_object(new_focus)
        focus_manager.get_manager().set_active_window(active_window)
        if old_focus is None:
            old_focus = active_window

        manager = presentation_manager.get_manager()
        manager.interrupt_if_needed_for_focus_change(old_focus, new_focus, event)

        if input_event_manager.get_manager().last_event_was_caret_selection():
            handled = False
            if AXObject.supports_text(old_focus):
                handled = (
                    text_selection_presenter.get_presenter().present_text_selection_change(
                        self,
                        old_focus,
                    )
                    or handled
                )
            if AXObject.supports_text(new_focus):
                handled = (
                    text_selection_presenter.get_presenter().present_text_selection_change(
                        self,
                        new_focus,
                    )
                    or handled
                )
            if handled:
                self.update_braille(new_focus)
                return True

        # A name change on the focused object usually means only its value changed (e.g.
        # arrowing a combo box); keep old_focus as prior_obj so we don't re-speak the
        # ancestor context. Reused table rows (e.g. Thunderbird message deletion) are the
        # exception: present those in full.
        prior = old_focus
        if is_name_change and old_focus == new_focus and AXUtilities.is_table_row(new_focus):
            prior = None
        manager.present_object(self, new_focus, prior_obj=prior)
        return True

    def activate(self) -> None:
        """Called when this script is activated."""

        tokens = ["DEFAULT: Activating script for", self.app]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if not focus_manager.get_manager().is_in_preferences_window():
            speech_manager.get_manager().update_punctuation_level()
            speech_manager.get_manager().update_capitalization_style()
            speech_manager.get_manager().update_synthesizer()

        presenter = document_presenter.get_presenter()
        if presenter.has_state_for_app(self.app):
            presenter.restore_mode_for_script(self)
        else:
            reason = "script activation, no prior state"
            presenter.suspend_navigators(self, False, reason)
            structural_navigator.get_navigator().set_mode(self, self._default_sn_mode)
            caret_navigator.get_navigator().set_enabled_for_script(
                self,
                self._default_caret_navigation_enabled,
            )

        table_navigator.get_navigator().refresh_enabled_state()
        command_manager.get_manager().activate_commands(f"activated {self.name}")

    def deactivate(self) -> None:
        """Called when this script is deactivated."""

        if bypass_mode_manager.get_manager().is_active():
            bypass_mode_manager.get_manager().toggle_enabled(self)

    def update_braille(
        self,
        obj: Atspi.Accessible,
        offset: int | None = None,
    ) -> None:
        """Updates the braille display to show obj."""

        if not obj:
            return
        braille_presenter.get_presenter().present_generated_braille(
            self,
            obj,
            offset=offset,
        )

    def handle_braille_pan_at_edge(
        self,
        _direction: braille_presenter.PanDirection,
    ) -> bool | None:
        """Handles braille panning when the presenter reaches the edge of a line."""

        return None

    ########################################################################
    #                                                                      #
    # AT-SPI OBJECT EVENT HANDLERS                                         #
    #                                                                      #
    # Event handlers return bool:                                          #
    # - True: Event was fully handled, no further processing needed        #
    # - False: Event wasn't handled, should be passed to parent handlers   #
    # Default return value is True (event handled by default script)       #
    #                                                                      #
    ########################################################################

    def _on_active_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:state-changed:active accessibility events."""

        window = event.source
        if AXUtilities.is_dialog_or_alert(window) or AXUtilities.is_frame(window):
            if event.detail1 and not AXUtilities.can_be_active_window(window):
                return True

            source_is_active_window = window == focus_manager.get_manager().get_active_window()
            if source_is_active_window and not event.detail1:
                focus = focus_manager.get_manager().get_locus_of_focus()
                if AXUtilities.is_menu_descendant(focus, inclusive=True):
                    msg = "DEFAULT: Ignoring event. In menu."
                    debug.print_message(debug.LEVEL_INFO, msg, True)
                    return True

                msg = "DEFAULT: Event is for active window. Clearing state."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                focus_manager.get_manager().set_active_window(None)
                return True

            if not source_is_active_window and event.detail1:
                msg = "DEFAULT: Updating active window."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                focus_manager.get_manager().set_active_window(
                    window,
                    set_window_as_focus=True,
                    notify_script=True,
                )

        return True

    def _on_active_descendant_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:active-descendant-changed accessibility events."""

        if AXUtilities.is_presentable_active_descendant_change(event):
            focus_manager.get_manager().set_locus_of_focus(event, event.any_data)

        return True

    def _on_checked_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:state-changed:checked accessibility events."""

        if AXUtilities.is_presentable_checked_change(event):
            presentation_manager.get_manager().interrupt_if_needed_for_object_presentation()
            self.present_object(event.source, reason=PresentationReason.STATE_CHANGE)

        return True

    def _on_children_added(self, event: Atspi.Event) -> bool:
        """Callback for object:children-changed:add accessibility events."""

        AXUtilities.clear_all_cache_now(event.source, "children-changed event.")
        return True

    def _on_children_removed(self, event: Atspi.Event) -> bool:
        """Callback for object:children-changed:remove accessibility events."""

        AXUtilities.clear_all_cache_now(event.source, "children-changed event.")
        return True

    # pylint: disable-next=too-many-return-statements
    def _on_caret_moved(self, event: Atspi.Event) -> bool:
        """Callback for object:text-caret-moved accessibility events."""

        reason = AXUtilities.get_text_event_reason(event)
        selection_reasons = {
            TextEventReason.SELECTION_BY_CHARACTER,
            TextEventReason.SELECTION_BY_LINE,
            TextEventReason.SELECTION_BY_PARAGRAPH,
            TextEventReason.SELECTION_BY_PAGE,
            TextEventReason.SELECTION_BY_WORD,
            TextEventReason.SELECTION_TO_FILE_BOUNDARY,
            TextEventReason.SELECTION_TO_LINE_BOUNDARY,
        }
        if (
            reason in selection_reasons
            and text_selection_manager.get_manager().get_current_selection_command(event.source)
            is not None
        ):
            msg = "DEFAULT: Ignoring caret-moved event caused by Orca's text selection"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if reason == TextEventReason.SEARCH_PRESENTABLE:
            contents = self.utilities.get_line_contents_at_offset(event.source, event.detail1)
            presentation_manager.get_manager().speak_contents(contents)
            msg = "DEFAULT: Presenting line for search result change"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if reason == TextEventReason.AUTO_INSERTION_UNPRESENTABLE:
            msg = "DEFAULT: Ignoring caret-moved event believed to be irrelevant auto insertion"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        manager = focus_manager.get_manager()
        focus = manager.get_locus_of_focus()
        if focus != event.source:
            if reason == TextEventReason.FOCUS_CHANGE:
                msg = "DEFAULT: Caret moved due to focus change (focus event missing)"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                manager.set_locus_of_focus(event, event.source)
                return True

            if not AXUtilities.is_focused(event.source):
                msg = "DEFAULT: Change is from unfocused source that is not the locus of focus"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return True
            # TODO - JD: See if this can be removed. If it's still needed document why.
            manager.set_locus_of_focus(event, event.source, False)

        if text_selection_manager.get_manager().get_current_selection_command(event.source):
            msg = "DEFAULT: Ignoring caret-moved event caused by Orca's text selection"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        obj, offset = manager.get_last_cursor_position()
        if offset == event.detail1 and obj == event.source:
            navigation_reasons = {
                TextEventReason.NAVIGATION_BY_WORD,
                TextEventReason.NAVIGATION_BY_CHARACTER,
                TextEventReason.NAVIGATION_BY_LINE,
                TextEventReason.NAVIGATION_TO_FILE_BOUNDARY,
            }
            if reason not in navigation_reasons:
                msg = "DEFAULT: Event is for last saved cursor position"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return True
            if caret_navigator.get_navigator().last_input_event_was_navigation_command():
                msg = "DEFAULT: Event is for position set by Orca's caret navigation"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return True
            msg = "DEFAULT: Position matches but proceeding due to navigation reason"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            presentation_manager.get_manager().interrupt_presentation()

        offset = AXText.get_caret_offset(event.source)

        # TODO - JD: These need to be harmonized / unified / simplified.
        manager.set_last_cursor_position(event.source, offset)
        self.utilities.set_caret_context(event.source, offset)

        # The text-changed and text-selection-changed handlers already refreshed braille
        # and presented speech for these reasons. Letting caret-moved also run would
        # re-emit the same braille line a second time for the same logical operation.
        ignore = [
            TextEventReason.BACKSPACE,
            TextEventReason.CUT,
            TextEventReason.DELETE,
            TextEventReason.PASTE,
            TextEventReason.REDO,
            TextEventReason.SPIN_BUTTON_VALUE_CHANGE,
            TextEventReason.UNDO,
        ]
        if reason in ignore:
            tokens = ["DEFAULT: Ignoring event due to reason (", reason, ")"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            AXUtilities.update_cached_selected_text(event.source)
            return True

        if AXUtilities.has_selected_text(event.source):
            msg = "DEFAULT: Event source has text selections"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            text_selection_presenter.get_presenter().present_text_selection_change(
                self,
                event.source,
            )
            return True

        text, _start, _end = AXUtilities.get_cached_selected_text(obj)
        if text and text_selection_presenter.get_presenter().present_text_selection_change(
            self,
            obj,
        ):
            msg = "DEFAULT: Event handled as text selection change"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        msg = "DEFAULT: Presenting text at new caret position"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        self._present_caret_moved_event(event, reason=reason)
        return True

    def _on_description_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:property-change:accessible-description events."""

        if AXUtilities.is_presentable_description_change(event):
            presentation_manager.get_manager().present_message(event.any_data)
        return True

    def _on_document_page_changed(self, event: Atspi.Event) -> bool:
        """Callback for document:page-changed accessibility events."""

        if event.detail1 < 0:
            return True

        if text_selection_manager.get_manager().defer_page_change_for_current_selection(
            event.detail1
        ):
            msg = "DEFAULT: Deferring page change caused by Orca's text selection"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        presentation_manager.get_manager().present_message(messages.PAGE_NUMBER % event.detail1)
        return True

    def _on_expanded_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:state-changed:expanded accessibility events."""

        AXUtilities.clear_all_cache_now(event.source, "expanded-changed event.")
        if not AXUtilities.is_presentable_expanded_change(event):
            return True

        presentation_manager.get_manager().interrupt_if_needed_for_object_presentation()
        self.present_object(event.source, reason=PresentationReason.STATE_CHANGE)
        details = AXUtilities.get_details_content(event.source)
        for detail in details:
            presentation_manager.get_manager().speak_message(detail)

        return True

    def _on_indeterminate_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:state-changed:indeterminate accessibility events."""

        if AXUtilities.is_presentable_indeterminate_change(event):
            presentation_manager.get_manager().interrupt_if_needed_for_object_presentation()
            self.present_object(event.source, reason=PresentationReason.STATE_CHANGE)

        return True

    def _on_invalid_entry_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:state-changed:invalid-entry accessibility events."""

        if not AXUtilities.is_presentable_invalid_entry_change(event):
            return True

        if event.detail1:
            msg = self.get_speech_generator().get_error_message(event.source)
        else:
            msg = messages.INVALID_ENTRY_FIXED
        presentation_manager.get_manager().speak_message(msg)
        self.update_braille(event.source)
        return True

    def _on_mouse_button(self, event: Atspi.Event) -> bool:
        """Callback for mouse:button events."""

        input_event_manager.get_manager().process_mouse_button_event(event)
        return True

    def _on_announcement(self, event: Atspi.Event) -> bool:
        """Callback for object:announcement events."""

        if isinstance(event.any_data, str):
            presentation_manager.get_manager().present_message(event.any_data)

        return True

    def _on_name_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:property-change:accessible-name events."""

        if not AXUtilities.is_presentable_name_change(event):
            return True

        manager = focus_manager.get_manager()
        if event.source == manager.get_locus_of_focus():
            if AXUtilities.is_combo_box(event.source):
                return True

            # Force the update so that braille is refreshed.
            manager.set_locus_of_focus(event, event.source, True, True)
            return True

        presentation_manager.get_manager().present_message(event.any_data)
        return True

    def _on_object_attributes_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:attributes-changed accessibility events."""

        AXUtilities.clear_all_cache_now(event.source, "object-attributes-changed event.")
        return True

    def _on_pressed_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:state-changed:pressed accessibility events."""

        if AXUtilities.is_presentable_pressed_change(event):
            presentation_manager.get_manager().interrupt_if_needed_for_object_presentation()
            self.present_object(event.source, reason=PresentationReason.STATE_CHANGE)

        return True

    def _on_selected_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:state-changed:selected accessibility events."""

        if not AXUtilities.is_presentable_selected_change(event):
            return True

        if speech_presenter.get_presenter().get_only_speak_displayed_text():
            return True

        announce_state = False
        manager = input_event_manager.get_manager()
        if manager.last_event_was_space():
            announce_state = True
        elif (
            manager.last_event_was_up() or manager.last_event_was_down()
        ) and AXUtilities.is_table_cell(event.source):
            announce_state = AXUtilities.is_selected(event.source)

        if not announce_state:
            return True

        # TODO - JD: Unlike the other state-changed callbacks, it seems unwise
        # to call generate_speech() here because that also will present the
        # expandable state if appropriate for the object type. The generators
        # need to gain some smarts w.r.t. state changes.

        if event.detail1:
            presentation_manager.get_manager().speak_message(messages.TEXT_SELECTED)
        else:
            presentation_manager.get_manager().speak_message(messages.TEXT_UNSELECTED)

        return True

    def _on_selection_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:selection-changed accessibility events."""

        if not AXUtilities.is_presentable_selection_change(event):
            return True

        presentation_manager.get_manager().present_command_announcement()

        if self.utilities.handle_container_selection_change(event.source):
            return True

        focus = focus_manager.get_manager().get_locus_of_focus()
        mouse_item = mouse_presenter.get_presenter().get_current_item()
        child = AXUtilities.get_selected_child_for_focus(
            event.source,
            focus,
            lambda c: c == mouse_item or AXUtilities.is_layout_only(c),
        )
        if AXUtilities.is_combo_box(event.source) and event.source == focus:
            self.present_object(event.source, reason=PresentationReason.STATE_CHANGE)
            return True

        if child is not None:
            if AXUtilities.is_page_tab(child) and not AXUtilities.is_focused(child):
                self.present_object(child, generate_braille=False)
            else:
                focus_manager.get_manager().set_locus_of_focus(event, child)

        return True

    def _on_focused_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:state-changed:focused accessibility events."""

        if not event.detail1:
            return True

        if not AXUtilities.is_focused(event.source):
            tokens = ["DEFAULT:", event.source, "lacks focused state. Clearing cache."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            AXObject.clear_cache(event.source, reason="Event detail1 does not match state.")
            if not AXUtilities.is_focused(event.source):
                msg = "DEFAULT: Clearing cache did not update state."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return True

        obj = event.source
        window, dialog = self.utilities.frame_and_dialog(obj)
        if window and not AXUtilities.can_be_active_window(window) and not dialog:
            return True

        if AXObject.get_child_count(obj) and not AXUtilities.is_combo_box(obj):
            selected_children = AXUtilities.selected_children(obj)
            if selected_children:
                obj = selected_children[0]

        focus_manager.get_manager().set_locus_of_focus(event, obj)
        return True

    def _on_showing_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:state-changed:showing accessibility events."""

        obj = event.source
        if AXUtilities.is_notification(obj):
            if not event.detail1:
                return True

            presentation_manager.get_manager().speak_message(
                self.get_speech_generator().get_localized_role_name(obj),
            )
            msg = self.utilities.get_notification_content(obj)
            presenter = presentation_manager.get_manager()
            presenter.speak_accessible_text(obj, msg)
            presenter.present_braille_message(msg)
            notification_presenter.get_presenter().save_notification(msg)
            return True

        if AXUtilities.is_tool_tip(obj):
            was_f1 = input_event_manager.get_manager().last_event_was_f1()
            if not was_f1 and not mouse_presenter.get_presenter().get_present_tooltips():
                return True
            if event.detail1:
                presentation_manager.get_manager().interrupt_if_needed_for_object_presentation()
                self.present_object(obj)
                return True

            focus = focus_manager.get_manager().get_locus_of_focus()
            if focus and was_f1:
                obj = focus
                presentation_manager.get_manager().interrupt_if_needed_for_object_presentation()
                self.present_object(obj, prior_obj=event.source)
                return True

        return True

    def _on_text_attributes_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:text-attributes-changed accessibility events."""

        if not AXUtilities.is_presentable_text_attributes_change(event):
            return True

        changes = AXUtilities.get_text_attribute_changes(event.source)

        word, start, _end = AXText.get_word_at_offset(event.source)
        if not word.strip():
            word, start, _end = AXText.get_word_at_offset(event.source, start - 1)

        speech_pres = speech_presenter.get_presenter()
        if error := speech_pres.get_error_description(event.source, start):
            presentation_manager.get_manager().speak_message(error)

        for attr, _old_value, new_value in changes:
            presentation_manager.get_manager().speak_message(attr.get_change_description(new_value))

        return True

    def _on_text_deleted(self, event: Atspi.Event) -> bool:
        """Callback for object:text-changed:delete accessibility events."""

        if not AXUtilities.is_presentable_text_deletion(event):
            return True

        reason = AXUtilities.get_text_event_reason(event)
        if reason == TextEventReason.AUTO_DELETION_UNPRESENTABLE:
            msg = "DEFAULT: Ignoring event believed to be irrelevant auto deletion"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if reason == TextEventReason.SPIN_BUTTON_VALUE_CHANGE:
            msg = "DEFAULT: Ignoring deletion; spin button value is presented via value-changed"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        presentation_manager.get_manager().present_command_announcement()
        self.update_braille(event.source)

        if reason == TextEventReason.SELECTED_TEXT_DELETION:
            msg = "DEFAULT: Deletion is believed to be due to deleting selected text"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            presentation_manager.get_manager().present_message(messages.SELECTION_DELETED)
            AXUtilities.update_cached_selected_text(event.source)
            return True

        text = self.utilities.deleted_text(event)
        selected_text, _start, _end = AXUtilities.get_cached_selected_text(event.source)
        if reason == TextEventReason.DELETE:
            msg = "DEFAULT: Deletion is believed to be due to Delete command"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            text = AXText.get_character_at_offset(event.source)[0]
        elif reason == TextEventReason.BACKSPACE and text != selected_text:
            msg = "DEFAULT: Deletion is believed to be due to BackSpace command"
            debug.print_message(debug.LEVEL_INFO, msg, True)
        else:
            msg = "DEFAULT: Event is not being presented due to lack of cause"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if len(text) == 1:
            presentation_manager.get_manager().speak_character(text)
        else:
            presentation_manager.get_manager().speak_accessible_text(event.source, text)

        return True

    def _on_text_inserted(self, event: Atspi.Event) -> bool:
        """Callback for object:text-changed:insert accessibility events."""

        if not AXUtilities.is_presentable_text_insertion(event):
            return True

        reason = AXUtilities.get_text_event_reason(event)
        if reason == TextEventReason.AUTO_INSERTION_UNPRESENTABLE:
            msg = "DEFAULT: Ignoring event believed to be irrelevant auto insertion"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if reason == TextEventReason.TERMINAL_LINE_NAVIGATION_REPAINT:
            msg = "DEFAULT: Presenting line due to terminal line-navigation repaint"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            offset = AXText.get_caret_offset(event.source)
            self.update_braille(event.source)
            self.say_line(event.source, offset)
            focus_manager.get_manager().set_last_cursor_position(event.source, offset)
            self.utilities.set_caret_context(event.source, offset)
            return True

        if reason == TextEventReason.SPIN_BUTTON_VALUE_CHANGE:
            msg = "DEFAULT: Ignoring insertion; spin button value is presented via value-changed"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        presentation_manager.get_manager().present_command_announcement()
        self.update_braille(event.source)

        if reason == TextEventReason.SELECTED_TEXT_RESTORATION:
            msg = "DEFAULT: Insertion is believed to be due to restoring selected text"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            presentation_manager.get_manager().present_message(messages.SELECTION_RESTORED)
            AXUtilities.update_cached_selected_text(event.source)
            return True

        if reason == TextEventReason.LIVE_REGION_UPDATE:
            msg = "DEFAULT: Event is from live region"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            live_region_presenter.get_presenter().handle_event(self, event)
            return True

        reason_messages: dict[TextEventReason, str] = {
            TextEventReason.PAGE_SWITCH: "due to page switch",
            TextEventReason.PASTE: "due to paste",
            TextEventReason.UNSPECIFIED_COMMAND: "due to command",
            TextEventReason.MOUSE_MIDDLE_BUTTON: "due to middle mouse button",
            TextEventReason.TYPING_ECHOABLE: "echoable",
            TextEventReason.AUTO_INSERTION_PRESENTABLE: "presentable auto text event",
            TextEventReason.SELECTED_TEXT_INSERTION: "also selected",
        }
        silent_reasons = {TextEventReason.PAGE_SWITCH, TextEventReason.PASTE}
        description = reason_messages.get(reason)
        if description is not None:
            tokens = ["DEFAULT: Insertion is believed to be", description]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            speak_string = reason not in silent_reasons
        else:
            msg = "DEFAULT: Not speaking inserted string due to lack of cause"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            speak_string = False

        # Because some implementations are broken.
        text = self.utilities.inserted_text(event)
        if speak_string:
            if len(text) == 1:
                presentation_manager.get_manager().speak_character(text)
            else:
                presentation_manager.get_manager().speak_accessible_text(event.source, text)

        if len(text) != 1 or reason not in [
            TextEventReason.TYPING,
            TextEventReason.TYPING_ECHOABLE,
        ]:
            return True

        if AXText.get_selected_ranges(event.source):
            return True

        presenter = typing_echo_presenter.get_presenter()
        if presenter.echo_previous_sentence(event.source):
            return True

        presenter.echo_previous_word(event.source)
        return True

    def _on_text_selection_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:text-selection-changed accessibility events."""

        selection_manager = text_selection_manager.get_manager()
        selection_command = selection_manager.get_selection_command_for_object(event.source)
        event_is_from_orca_selection = selection_command is not None
        if (
            AXUtilities.is_focusable(event.source)
            and not AXUtilities.is_focused(event.source)
            and not event_is_from_orca_selection
        ):
            msg = "DEFAULT: Ignoring event from focusable but unfocused source"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        # We won't handle undo here as it can lead to double-presentation.
        # If there is an application for which text-changed events are
        # missing upon undo, handle them in an app or toolkit script.

        reason = AXUtilities.get_text_event_reason(event)
        if reason == TextEventReason.UNKNOWN and not event_is_from_orca_selection:
            msg = "DEFAULT: Ignoring event because reason for change is unknown"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            AXUtilities.update_cached_selected_text(event.source)
            return True
        if reason == TextEventReason.SPIN_BUTTON_VALUE_CHANGE:
            msg = "DEFAULT: Ignoring selection; spin button value is presented via value-changed"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            AXUtilities.update_cached_selected_text(event.source)
            return True
        if reason == TextEventReason.SEARCH_PRESENTABLE:
            msg = "DEFAULT: Presenting line believed to be search match"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self.say_line(event.source, AXText.get_caret_offset(event.source))
            AXUtilities.update_cached_selected_text(event.source)
            return True
        if reason == TextEventReason.SEARCH_UNPRESENTABLE:
            msg = "DEFAULT: Ignoring event believed to be unpresentable search results change"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            AXUtilities.update_cached_selected_text(event.source)
            return True
        if reason in [TextEventReason.CUT, TextEventReason.BACKSPACE, TextEventReason.DELETE]:
            msg = "DEFAULT: Ignoring event believed to be text removal"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            AXUtilities.update_cached_selected_text(event.source)
            return True
        if reason == TextEventReason.AUTO_SELECTION:
            msg = "DEFAULT: Ignoring auto-selection believed to be side effect of typing"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            AXUtilities.update_cached_selected_text(event.source)
            return True
        if reason == TextEventReason.AUTO_UNSELECTION:
            msg = "DEFAULT: Ignoring auto-unselection believed to be side effect of typing"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            AXUtilities.update_cached_selected_text(event.source)
            return True

        selection_objs = (event.source,)
        if selection_command is not None:
            selection_objs = selection_command.get_objects() or selection_objs
            container = selection_command.get_selection_container()
            if container is not None and AXUtilities.is_document(container):
                selection_objs = (event.source,)
        tokens = ["DEFAULT: Reconciling text selection changes in", selection_objs]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        for selection_obj in selection_objs:
            if reason == TextEventReason.SELECTION_UNPRESENTABLE:
                selection_manager.update_state_for_unpresentable_selection_change(selection_obj)
            else:
                text_selection_presenter.get_presenter().present_text_selection_change(
                    self,
                    selection_obj,
                )
        self.update_braille(event.source)
        return True

    def _on_column_reordered(self, event: Atspi.Event) -> bool:
        """Callback for object:column-reordered accessibility events."""

        AXUtilities.clear_all_cache_now(event.source, "column-reordered event.")
        if not input_event_manager.get_manager().last_event_was_table_sort():
            return True

        if event.source != AXUtilities.get_table(focus_manager.get_manager().get_locus_of_focus()):
            return True

        presentation_manager.get_manager().present_message(messages.TABLE_REORDERED_COLUMNS)
        return True

    def _on_row_reordered(self, event: Atspi.Event) -> bool:
        """Callback for object:row-reordered accessibility events."""

        AXUtilities.clear_all_cache_now(event.source, "row-reordered event.")
        if not input_event_manager.get_manager().last_event_was_table_sort():
            return True

        if event.source != AXUtilities.get_table(focus_manager.get_manager().get_locus_of_focus()):
            return True

        presentation_manager.get_manager().present_message(messages.TABLE_REORDERED_ROWS)
        return True

    def _on_value_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:property-change:accessible-value accessibility events."""

        if not AXUtilities.is_presentable_value_change(event):
            return True

        manager = focus_manager.get_manager()
        if (
            not AXUtilities.is_progress_bar(event.source)
            and event.source != manager.get_locus_of_focus()
        ):
            msg = "DEFAULT: Source != locusOfFocus"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if AXUtilities.is_spin_button(event.source):
            manager.set_last_cursor_position(event.source, AXText.get_caret_offset(event.source))

        is_progress_bar = AXUtilities.is_progress_bar(event.source)
        if not is_progress_bar:
            presentation_manager.get_manager().interrupt_presentation()

        presentation_manager.get_manager().present_object(
            self,
            event.source,
            generate_sound=True,
            reason=(
                PresentationReason.PROGRESS_BAR_UPDATE
                if is_progress_bar
                else PresentationReason.STATE_CHANGE
            ),
        )
        return True

    def _on_window_activated(self, event: Atspi.Event) -> bool:
        """Callback for window:activate accessibility events."""

        if not AXUtilities.can_be_active_window(event.source):
            return True

        manager = focus_manager.get_manager()
        active_window = manager.get_active_window()
        if event.source == active_window:
            msg = "DEFAULT: Event is for active window."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            spellcheck_presenter.get_presenter().handle_window_event(event, self)
            return True

        manager.set_active_window(event.source)
        if AXUtilities.is_combo_box_popup(active_window):
            return True

        if AXObject.get_child_count(event.source) == 1:
            child = AXObject.get_child(event.source, 0)
            # Popup menus in Chromium live in a menu bar whose first child is a panel.
            if AXUtilities.is_menu_bar(child):
                child = AXUtilities.get_menu(child)
            if AXUtilities.is_menu(child):
                selected_item = AXSelection.get_selected_child(child, 0)
                if AXUtilities.is_selected(selected_item):
                    child = selected_item
                manager.set_locus_of_focus(event, child)
                return True

        if not spellcheck_presenter.get_presenter().handle_window_event(event, self):
            manager.set_locus_of_focus(event, event.source)
        return True

    def _on_window_deactivated(self, event: Atspi.Event) -> bool:
        """Callback for window:deactivate accessibility events."""

        manager = focus_manager.get_manager()
        focus = manager.get_locus_of_focus()
        if AXUtilities.is_menu_descendant(focus, inclusive=True):
            msg = "DEFAULT: Ignoring event. In menu."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if event.source != manager.get_active_window():
            msg = "DEFAULT: Ignoring event. Not for active window"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if AXUtilities.is_combo_box(focus) and AXUtilities.is_combo_box_popup(
            AXUtilities.find_active_window()
        ):
            msg = "DEFAULT: Ignoring event. Combo box popup is the new active window."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if AXUtilities.is_combo_box_popup(event.source):
            msg = "DEFAULT: Ignoring event. Combo box popup."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if learn_mode_presenter.get_presenter().is_active():
            learn_mode_presenter.get_presenter().quit()

        focus_manager.get_manager().clear_state("Window deactivated")
        script_manager.get_manager().set_active_script(None, "Window deactivated")
        return True

    ########################################################################
    #                                                                      #
    # Methods for presenting content                                       #
    #                                                                      #
    ########################################################################

    def _update_braille_caret_position(self, obj: Atspi.Accessible) -> None:
        """Try to reposition the cursor without having to do a full update."""

        if not braille_presenter.get_presenter().use_braille():
            return

        if braille.try_reposition_cursor(obj):
            return

        self.update_braille(obj)

    def _present_caret_moved_event(
        self,
        event: Atspi.Event,
        obj: Atspi.Accessible | None = None,
        reason: TextEventReason = TextEventReason.UNKNOWN,
    ) -> bool:
        """Presents text at the new position, based on heuristics. Returns True if handled."""

        obj = obj or event.source
        self._update_braille_caret_position(obj)

        # TODO - JD: Make this a TextEventReason. Also handle structural navigation
        # and table navigation here. Technically that's not been necessary because
        # it won't match anything below. But it would be cleaner to cover each cause
        # explicitly.
        if caret_navigator.get_navigator().last_input_event_was_navigation_command():
            msg = "DEFAULT: Event ignored: Last command was caret nav"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if reason == TextEventReason.SAY_ALL:
            msg = "DEFAULT: Not presenting text because SayAll is active"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        line_reasons = (
            TextEventReason.NAVIGATION_BY_LINE,
            TextEventReason.NAVIGATION_BY_PAGE,
            TextEventReason.NAVIGATION_TO_FILE_BOUNDARY,
        )
        if reason in line_reasons:
            self.say_line(obj, event.detail1)
            return True

        if reason == TextEventReason.NAVIGATION_BY_WORD:
            self.say_word(obj, event.detail1)
            return True

        character_reasons = (
            TextEventReason.NAVIGATION_BY_CHARACTER,
            TextEventReason.NAVIGATION_TO_LINE_BOUNDARY,
        )
        if reason in character_reasons:
            self.say_character(obj, event.detail1)
            return True

        if reason == TextEventReason.MOUSE_PRIMARY_BUTTON:
            text, _start, _end = AXUtilities.get_cached_selected_text(event.source)
            if not text:
                self.say_line(obj, event.detail1)
                return True
        return False

    def say_character(self, obj: Atspi.Accessible, offset: int) -> None:
        """Speak the character at the specified offset."""

        if input_event_manager.get_manager().last_event_was_forward_caret_selection():
            offset -= 1

        character, start_offset, end_offset = AXText.get_character_at_offset(
            obj, offset, ensure_whole_characters=True
        )
        focus_manager.get_manager().emit_region_changed(
            obj,
            start_offset,
            end_offset,
            focus_manager.CARET_TRACKING,
        )

        presentation_manager.get_manager().speak_character_at_offset(obj, offset, character)
        AXUtilities.set_last_text_unit_spoken(TextUnit.CHAR)

    def say_line(self, obj: Atspi.Accessible, offset: int) -> None:
        """Speaks the line at the specified offset."""

        line, start_offset = AXText.get_line_at_offset(obj, offset)[0:2]
        if line and line != "\n":
            end_offset = start_offset + len(line)
            focus_manager.get_manager().emit_region_changed(
                obj,
                start_offset,
                end_offset,
                focus_manager.CARET_TRACKING,
            )

        speech_presenter.get_presenter().speak_line(
            self,
            obj,
            start_offset,
            start_offset + len(line),
            line,
        )

        AXUtilities.set_last_text_unit_spoken(TextUnit.LINE)

    def say_phrase(self, obj: Atspi.Accessible, start_offset: int, end_offset: int) -> None:
        """Speaks the substring between start and end offset."""

        phrase = AXUtilities.expand_eocs(obj, start_offset, end_offset)
        if not phrase:
            return

        if len(phrase) > 1 or phrase.isalnum():
            focus_manager.get_manager().emit_region_changed(
                obj,
                start_offset,
                end_offset,
                focus_manager.CARET_TRACKING,
            )

        speech_presenter.get_presenter().speak_phrase(self, obj, start_offset, end_offset, phrase)
        AXUtilities.set_last_text_unit_spoken(TextUnit.PHRASE)

    def say_word(self, obj: Atspi.Accessible, offset: int) -> None:
        """Speaks the word at the specified offset, taking into account the previous position."""

        word, start_offset, end_offset = self.utilities.get_word_at_offset_adjusted_for_navigation(
            obj,
            offset,
        )
        if not word and offset == AXText.get_character_count(obj):
            self.say_character(obj, offset)
            AXUtilities.set_last_text_unit_spoken(TextUnit.WORD)
            return

        speech_pres = speech_presenter.get_presenter()
        if "\n" in word:
            # Announce when we cross a hard line boundary, based on whether or not indentation
            # should be spoken. This was done to avoid yet another setting in response to some
            # users saying this announcement was too chatty. The idea of using this setting for
            # the decision is that if the user wants indentation announced, they are interested
            # in explicit whitespace information.
            if speech_pres.get_speak_indentation():
                presentation_manager.get_manager().speak_character("\n")
            if word.startswith("\n"):
                start_offset += 1
            elif word.endswith("\n"):
                end_offset -= 1
            word = AXText.get_substring(obj, start_offset, end_offset)

        # say_phrase is useful because it handles punctuation verbalization, but we don't want
        # to trigger its whitespace presentation.
        matches = list(re.finditer(r"\S+", word))
        if matches:
            start_offset += matches[0].start()
            end_offset -= len(word) - matches[-1].end()
            word = AXText.get_substring(obj, start_offset, end_offset)

        text = word.replace("\n", "\\n")
        tokens = [
            "DEFAULT: Final word at offset",
            offset,
            "is '",
            text,
            "' (",
            start_offset,
            "-",
            end_offset,
            ")",
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        speech_presenter.get_presenter().present_text_attribute_state(obj, start_offset)
        self.say_phrase(obj, start_offset, end_offset)
        AXUtilities.set_last_text_unit_spoken(TextUnit.WORD)

    def present_object(
        self,
        obj: Atspi.Accessible,
        offset: int | None = None,
        prior_obj: Atspi.Accessible | None = None,
        generate_speech: bool = True,
        generate_braille: bool = True,
        reason: PresentationReason | None = None,
    ) -> None:
        """Presents the current object."""

        tokens = ["DEFAULT: Presenting object", obj]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if offset is not None:
            AXUtilities.set_caret_offset_with_reason(
                obj, offset, CaretSetReason.OBJECT_PRESENTATION
            )
            if text_selection_manager.get_manager().has_known_selection(obj):
                text_selection_presenter.get_presenter().present_text_selection_change(
                    self,
                    obj,
                )

        presentation_manager.get_manager().present_object(
            self,
            obj,
            generate_speech=generate_speech,
            generate_braille=generate_braille,
            prior_obj=prior_obj,
            reason=reason,
        )
